import io
import textwrap
from typing import Callable, Optional, List, Tuple, Sequence

import pytest

from debputy.commands.debputy_cmd.output import no_fancy_output
from debputy.dh_migration.migrators import Migrator
from debputy.dh_migration.migrators_impl import (
    migrate_tmpfile,
    migrate_lintian_overrides_files,
    detect_pam_files,
    migrate_doc_base_files,
    migrate_installexamples_file,
    migrate_installdocs_file,
    migrate_install_file,
    migrate_maintscript,
    migrate_links_files,
    detect_dh_addons_with_zz_integration,
    migrate_not_installed_file,
    migrate_installman_file,
    migrate_bash_completion,
    migrate_installinfo_file,
    migrate_dh_installsystemd_files,
    detect_obsolete_substvars,
    detect_dh_addons_zz_debputy_rrr,
    migrate_shell_completions,
)
from debputy.dh_migration.models import (
    FeatureMigration,
    AcceptableMigrationIssues,
    UnsupportedFeature,
)
from debputy.highlevel_manifest import HighLevelManifest
from debputy.highlevel_manifest_parser import YAMLManifestParser
from debputy.plugin.api import virtual_path_def, VirtualPath
from debputy.plugin.api.spec import (
    INTEGRATION_MODE_DH_DEBPUTY_RRR,
    INTEGRATION_MODE_DH_DEBPUTY,
)
from debputy.plugin.api.test_api import (
    build_virtual_file_system,
)


DEBIAN_DIR_ENTRY = virtual_path_def(".", fs_path="/nowhere/debian")


@pytest.fixture()
def manifest_parser_pkg_foo_factory(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
    source_package,
    package_single_foo_arch_all_cxt_amd64,
    amd64_substitution,
    no_profiles_or_build_options,
    debputy_plugin_feature_set,
) -> Callable[[], YAMLManifestParser]:
    # We need an empty directory to avoid triggering packager provided files.
    debian_dir = build_virtual_file_system([])

    def _factory():
        return YAMLManifestParser(
            "debian/test-debputy.manifest",
            source_package,
            package_single_foo_arch_all_cxt_amd64,
            amd64_substitution,
            amd64_dpkg_architecture_variables,
            dpkg_arch_query,
            no_profiles_or_build_options,
            debputy_plugin_feature_set,
            "full",
            debian_dir=debian_dir,
        )

    return _factory


@pytest.fixture(scope="session")
def accept_no_migration_issues() -> AcceptableMigrationIssues:
    return AcceptableMigrationIssues(frozenset())


@pytest.fixture(scope="session")
def accept_any_migration_issues() -> AcceptableMigrationIssues:
    return AcceptableMigrationIssues(frozenset(["ALL"]))


@pytest.fixture
def empty_manifest_pkg_foo(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
) -> HighLevelManifest:
    return manifest_parser_pkg_foo_factory().build_manifest()


def run_migrator(
    migrator: Migrator,
    path: VirtualPath,
    manifest: HighLevelManifest,
    acceptable_migration_issues: AcceptableMigrationIssues,
    *,
    migration_target=INTEGRATION_MODE_DH_DEBPUTY,
) -> FeatureMigration:
    feature_migration = FeatureMigration(migrator.__name__, no_fancy_output())
    migrator(
        path,
        manifest,
        acceptable_migration_issues,
        feature_migration,
        migration_target,
    )
    return feature_migration


def _assert_unsupported_feature(
    migrator: Migrator,
    path: VirtualPath,
    manifest: HighLevelManifest,
    acceptable_migration_issues: AcceptableMigrationIssues,
):
    with pytest.raises(UnsupportedFeature) as e:
        run_migrator(migrator, path, manifest, acceptable_migration_issues)
    return e


def _write_manifest(manifest: HighLevelManifest) -> str:
    mutable_manifest = manifest.mutable_manifest
    assert mutable_manifest is not None
    with io.StringIO() as fd:
        mutable_manifest.write_to(fd)
        return fd.getvalue()


def _verify_migrator_generates_parsable_manifest(
    migrator: Migrator,
    parser_factory: Callable[[], YAMLManifestParser],
    acceptable_migration_issues: AcceptableMigrationIssues,
    dh_config_name: str,
    dh_config_content: str,
    expected_manifest_content: str,
    expected_warnings: Optional[List[str]] = None,
    expected_renamed_paths: Optional[List[Tuple[str, str]]] = None,
    expected_removals: Optional[List[str]] = None,
    required_plugins: Sequence[str] = tuple(),
    dh_config_mode: Optional[int] = None,
) -> None:
    # No file, no changes
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    migration = run_migrator(
        migrator,
        empty_fs,
        parser_factory().build_manifest(),
        acceptable_migration_issues,
    )

    assert not migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    if dh_config_mode is None:
        if dh_config_content.startswith(("#!/usr/bin/dh-exec", "#! /usr/bin/dh-exec")):
            dh_config_mode = 0o755
        else:
            dh_config_mode = 0o644

    # Test with a dh_config file now
    fs_w_dh_config = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                dh_config_name,
                fs_path=f"/nowhere/debian/{dh_config_name}",
                content=dh_config_content,
                mode=dh_config_mode,
            ),
        ]
    )
    manifest = parser_factory().build_manifest()

    migration = run_migrator(
        migrator,
        fs_w_dh_config,
        manifest,
        acceptable_migration_issues,
    )

    assert migration.anything_to_do
    if expected_warnings is not None:
        assert migration.warnings == expected_warnings
    else:
        assert not migration.warnings
    assert migration.remove_paths_on_success == [f"/nowhere/debian/{dh_config_name}"]
    if expected_removals is None:
        assert migration.remove_paths_on_success == [
            f"/nowhere/debian/{dh_config_name}"
        ]
    else:
        assert migration.remove_paths_on_success == expected_removals
    if expected_renamed_paths is not None:
        assert migration.rename_paths_on_success == expected_renamed_paths
    else:
        assert not migration.rename_paths_on_success
    assert tuple(migration.required_plugins) == tuple(required_plugins)
    actual_manifest = _write_manifest(manifest)
    assert actual_manifest == expected_manifest_content

    # Verify that the output is actually parsable
    parser_factory().parse_manifest(fd=actual_manifest)


def test_migrate_tmpfile(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = migrate_tmpfile
    empty_debian_dir = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    migration = run_migrator(
        migrator,
        empty_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert not migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success

    tmpfile_debian_dir = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("tmpfile", fs_path="/nowhere/debian/tmpfile"),
        ]
    )

    migration = run_migrator(
        migrator,
        tmpfile_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert migration.rename_paths_on_success == [
        ("/nowhere/debian/tmpfile", "/nowhere/debian/tmpfiles")
    ]

    tmpfile_debian_dir = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            # Use real paths here to make `cmp -s` discover that they are the same
            virtual_path_def("tmpfile", fs_path="debian/control"),
            virtual_path_def("tmpfiles", fs_path="debian/control"),
        ]
    )

    migration = run_migrator(
        migrator,
        tmpfile_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    assert migration.anything_to_do
    assert not migration.warnings
    assert migration.remove_paths_on_success == ["debian/control"]
    assert not migration.rename_paths_on_success

    conflict_tmpfile_debian_dir = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            # Use real paths here to make `cmp -s` discover a difference
            virtual_path_def("tmpfile", fs_path="debian/control"),
            virtual_path_def("tmpfiles", fs_path="debian/changelog"),
        ]
    )

    migration = run_migrator(
        migrator,
        conflict_tmpfile_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert len(migration.warnings) == 1
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success

    conflict_tmpfile_debian_dir = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("tmpfile", fs_path="/nowhere/debian/tmpfile"),
            virtual_path_def("tmpfiles/", fs_path="/nowhere/debian/tmpfiles"),
        ]
    )

    migration = run_migrator(
        migrator,
        conflict_tmpfile_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert len(migration.warnings) == 1
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success

    conflict_tmpfile_debian_dir = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "tmpfile",
                link_target="/nowhere/debian/tmpfiles",
                fs_path="/nowhere/debian/tmpfile",
            ),
        ]
    )

    migration = run_migrator(
        migrator,
        conflict_tmpfile_debian_dir,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert len(migration.warnings) == 1
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success


def test_migrate_lintian_overrides_files(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
    accept_any_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = migrate_lintian_overrides_files
    no_override_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    single_noexec_override_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "foo.lintian-overrides",
                fs_path="/nowhere/no-exec/debian/foo.lintian-overrides",
            ),
        ]
    )
    single_exec_override_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "foo.lintian-overrides",
                fs_path="/nowhere/exec/debian/foo.lintian-overrides",
                mode=0o755,
            ),
        ]
    )
    for no_issue_fs in [no_override_fs, single_noexec_override_fs]:
        migration = run_migrator(
            migrator,
            no_issue_fs,
            empty_manifest_pkg_foo,
            accept_no_migration_issues,
        )

        assert not migration.anything_to_do
        assert not migration.warnings
        assert not migration.remove_paths_on_success
        assert not migration.rename_paths_on_success

    _assert_unsupported_feature(
        migrator,
        single_exec_override_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    migration = run_migrator(
        migrator,
        single_exec_override_fs,
        empty_manifest_pkg_foo,
        accept_any_migration_issues,
    )

    assert migration.anything_to_do
    assert len(migration.warnings) == 1
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success


def test_detect_pam_files(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_pam_files
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    pam_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("pam", fs_path="/nowhere/debian/foo.pam"),
        ]
    )

    migration = run_migrator(
        migrator,
        empty_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert not migration.anything_to_do
    assert not migration.warnings
    assert migration.assumed_compat is None
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success

    migration = run_migrator(
        migrator,
        pam_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert not migration.anything_to_do
    assert not migration.warnings
    assert migration.assumed_compat == 14
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success


def test_migrate_doc_base_files(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = migrate_doc_base_files
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    doc_base_ok_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("doc-base", fs_path="/nowhere/debian/doc-base"),
            virtual_path_def(
                "foo.doc-base.EX", fs_path="/nowhere/debian/foo.doc-base.EX"
            ),
        ]
    )
    doc_base_migration_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "foo.doc-base.bar", fs_path="/nowhere/debian/foo.doc-base.bar"
            ),
        ]
    )

    for no_change_fs in [empty_fs, doc_base_ok_fs]:
        migration = run_migrator(
            migrator,
            no_change_fs,
            empty_manifest_pkg_foo,
            accept_no_migration_issues,
        )

        assert not migration.anything_to_do
        assert not migration.warnings
        assert not migration.remove_paths_on_success
        assert not migration.rename_paths_on_success

    migration = run_migrator(
        migrator,
        doc_base_migration_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert migration.rename_paths_on_success == [
        ("/nowhere/debian/foo.doc-base.bar", "/nowhere/debian/foo.bar.doc-base")
    ]


def test_migrate_dh_installsystemd_files(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = migrate_dh_installsystemd_files
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])
    files_ok_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("@service", fs_path="/nowhere/debian/@service"),
            virtual_path_def("foo.@service", fs_path="/nowhere/debian/foo.@service"),
        ]
    )
    migration_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def("foo@.service", fs_path="/nowhere/debian/foo@.service"),
        ]
    )

    for no_change_fs in [empty_fs, files_ok_fs]:
        migration = run_migrator(
            migrator,
            no_change_fs,
            empty_manifest_pkg_foo,
            accept_no_migration_issues,
        )

        assert not migration.anything_to_do
        assert not migration.warnings
        assert not migration.remove_paths_on_success
        assert not migration.rename_paths_on_success

    migration = run_migrator(
        migrator,
        migration_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert migration.rename_paths_on_success == [
        ("/nowhere/debian/foo@.service", "/nowhere/debian/foo.@service")
    ]


def test_migrate_installexamples_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
        bar
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-examples:
                sources:
                - foo/*
                - bar
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installexamples_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "examples",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_installinfo_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
        bar
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-docs:
                sources:
                - foo/*
                - bar
                dest-dir: '{{path:GNU_INFO_DIR}}'
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installinfo_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "info",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_installinfo_file_conditionals(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        #!/usr/bin/dh-exec
        foo/* <!pkg.foo.noinfo>
        bar <!pkg.foo.noinfo>
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-docs:
                sources:
                - foo/*
                - bar
                dest-dir: '{{path:GNU_INFO_DIR}}'
                when:
                  build-profiles-matches: <!pkg.foo.noinfo>
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installinfo_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "info",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_installexamples_file_single_source(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-examples:
                source: foo/*
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installexamples_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "examples",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_installdocs_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
        bar
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-docs:
                sources:
                - foo/*
                - bar
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installdocs_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "docs",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_installdocs_file_single_source(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-docs:
                source: foo/*
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_installdocs_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "docs",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        bar usr/bin
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                source: bar
                dest-dir: usr/bin
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file_conditionals_simple_arch(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        #!/usr/bin/dh-exec
        bar usr/bin  [linux-any]
        foo usr/bin  [linux-any]
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - bar
                - foo
                dest-dir: usr/bin
                when:
                  arch-matches: linux-any
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file_util_linux_locales(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    # Parts of the `d/util-linux-locales.install` file. It uses d/tmp for most of its paths
    # and that breaks the default dest-dir (dh_install always strips --sourcedir, `debputy`
    # currently does not)
    dh_config_content = textwrap.dedent(
        """\
        #!/usr/bin/dh-exec
        usr/share/locale/*/*/util-linux.mo

        # bsdextrautils
        debian/tmp/usr/share/man/*/man1/col.1 <!nodoc>

        debian/tmp/usr/share/man/*/man3/libblkid.3 <!nodoc>
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - usr/share/man/*/man1/col.1
                - usr/share/man/*/man3/libblkid.3
                when:
                  build-profiles-matches: <!nodoc>
            - install:
                source: usr/share/locale/*/*/util-linux.mo
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file_conditionals_simple_combined_cond(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    for cond in ["<!foo> <!bar> [linux-any]", "[linux-any] <!foo> <!bar>"]:
        dh_config_content = textwrap.dedent(
            """\
            #!/usr/bin/dh-exec
            bar usr/bin  {CONDITION}
            foo usr/bin  {CONDITION}
        """
        ).format(CONDITION=cond)
        expected_manifest_content = textwrap.dedent(
            """\
                manifest-version: '0.1'
                installations:
                - install:
                    sources:
                    - bar
                    - foo
                    dest-dir: usr/bin
                    when:
                      all-of:
                      - arch-matches: linux-any
                      - build-profiles-matches: <!foo> <!bar>
        """
        )
        _verify_migrator_generates_parsable_manifest(
            migrate_install_file,
            manifest_parser_pkg_foo_factory,
            accept_no_migration_issues,
            "install",
            dh_config_content,
            expected_manifest_content,
        )


def test_migrate_install_file_conditionals_unknown_subst(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_any_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        #!/usr/bin/dh-exec
        bar ${unknown_substvar}
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            definitions:
              variables:
                unknown_substvar: 'TODO: Provide variable value for unknown_substvar'
            installations:
            - install:
                source: bar
                dest-dir: '{{unknown_substvar}}'
    """
    )
    expected_warning = (
        "TODO: MANUAL MIGRATION of unresolved substitution variable {{unknown_substvar}}"
        ' from ./install line 2 token "${unknown_substvar}"'
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_any_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
        expected_warnings=[expected_warning],
    )


def test_migrate_install_file_multidest(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        # Issue #66
        # - observed in kafs-client / the original install file copied in here.

        src/aklog-kafs        usr/bin
        src/kafs-check-config usr/bin
        #
        src/kafs-preload usr/sbin
        #
        src/kafs-dns     usr/libexec
        #
        conf/cellservdb.conf usr/share/kafs-client
        conf/client.conf     etc/kafs
        #
        conf/kafs_dns.conf etc/request-key.d
        #
        conf/cellservdb.conf usr/share/kafs
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - src/aklog-kafs
                - src/kafs-check-config
                dest-dir: usr/bin
            - install:
                source: src/kafs-preload
                dest-dir: usr/sbin
            - install:
                source: src/kafs-dns
                dest-dir: usr/libexec
            - install:
                source: conf/client.conf
                dest-dir: etc/kafs
            - install:
                source: conf/kafs_dns.conf
                dest-dir: etc/request-key.d
            - multi-dest-install:
                source: conf/cellservdb.conf
                dest-dirs:
                - usr/share/kafs-client
                - usr/share/kafs
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file_multidest_default_dest(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        # Relaed to issue #66 - testing corner case not present in the original install file

        src/aklog-kafs        usr/bin
        src/kafs-check-config usr/bin
        #
        src/kafs-preload usr/sbin
        #
        src/kafs-dns     usr/libexec
        #
        usr/share/kafs-client/cellservdb.conf
        conf/client.conf     etc/kafs
        #
        conf/kafs_dns.conf etc/request-key.d
        #
        usr/share/kafs-client/cellservdb.conf usr/share/kafs
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - src/aklog-kafs
                - src/kafs-check-config
                dest-dir: usr/bin
            - install:
                source: src/kafs-preload
                dest-dir: usr/sbin
            - install:
                source: src/kafs-dns
                dest-dir: usr/libexec
            - install:
                source: conf/client.conf
                dest-dir: etc/kafs
            - install:
                source: conf/kafs_dns.conf
                dest-dir: etc/request-key.d
            - multi-dest-install:
                source: usr/share/kafs-client/cellservdb.conf
                dest-dirs:
                - usr/share/kafs
                - usr/share/kafs-client
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_install_file_multidest_default_dest_warning(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        # Relaed to issue #66 - testing corner case not present in the original install file

        src/aklog-kafs        usr/bin
        src/kafs-check-config usr/bin
        #
        src/kafs-preload usr/sbin
        #
        src/kafs-dns     usr/libexec
        #
        usr/share/kafs-*/cellservdb.conf
        conf/client.conf     etc/kafs
        #
        conf/kafs_dns.conf etc/request-key.d
        #
        usr/share/kafs-*/cellservdb.conf usr/share/kafs
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - src/aklog-kafs
                - src/kafs-check-config
                dest-dir: usr/bin
            - install:
                source: src/kafs-preload
                dest-dir: usr/sbin
            - install:
                source: src/kafs-dns
                dest-dir: usr/libexec
            - install:
                source: conf/client.conf
                dest-dir: etc/kafs
            - install:
                source: conf/kafs_dns.conf
                dest-dir: etc/request-key.d
            - multi-dest-install:
                source: usr/share/kafs-*/cellservdb.conf
                dest-dirs:
                - usr/share/kafs
                - 'FIXME: usr/share/kafs-* (could not reliably compute the dest dir)'
    """
    )
    warnings = [
        "TODO: FIXME left in dest-dir(s) of some installation rules."
        " Please review these and remove the FIXME (plus correct as necessary)"
    ]
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
        expected_warnings=warnings,
    )


def test_migrate_installman_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        man/foo.1 man/bar.1
        man2/*.2
        man3/bar.3 man3/bar.de.3
        man/de/man3/bar.pl.3
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install-man:
                sources:
                - man/foo.1
                - man/bar.1
                - man2/*.2
                - man/de/man3/bar.pl.3
            - install-man:
                sources:
                - man3/bar.3
                - man3/bar.de.3
                language: derive-from-basename
    """
    )
    expected_warnings = [
        'Detected man pages that might rely on "derive-from-basename" logic.  Please double check'
        " that the generated `install-man` rules are correct"
    ]
    _verify_migrator_generates_parsable_manifest(
        migrate_installman_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "manpages",
        dh_config_content,
        expected_manifest_content,
        expected_warnings=expected_warnings,
    )


def test_migrate_install_dh_exec_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        #!/usr/bin/dh-exec

        foo/script.sh => usr/bin/script
        => usr/bin/bar
        usr/bin/* usr/share/foo/extra/* usr/share/foo/extra
        another-util usr/share/foo/extra
        # This will not be merged with `=> usr/bin/bar`
        usr/share/foo/features
        usr/share/foo/bugs
        some-file.txt usr/share/foo/online-doc
        # TODO: Support migration of these
        # pathA pathB  conditional/arch [linux-anx]
        # <!pkg.foo.condition>  another-path conditional/profile
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                source: usr/bin/bar
            - install:
                source: foo/script.sh
                as: usr/bin/script
            - install:
                sources:
                - usr/bin/*
                - usr/share/foo/extra/*
                - another-util
                dest-dir: usr/share/foo/extra
            - install:
                source: some-file.txt
                dest-dir: usr/share/foo/online-doc
            - install:
                sources:
                - usr/share/foo/features
                - usr/share/foo/bugs
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_install_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "install",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_maintscript(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        rm_conffile /etc/foo.conf
        mv_conffile /etc/bar.conf /etc/new-foo.conf 1.0~ bar
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            packages:
              foo:
                conffile-management:
                - remove:
                    path: /etc/foo.conf
                - rename:
                    source: /etc/bar.conf
                    target: /etc/new-foo.conf
                    prior-to-version: 1.0~
                    owning-package: bar
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_maintscript,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "maintscript",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_not_installed_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*.txt bar/${DEB_HOST_MULTIARCH}/*.so*
        baz/script.sh
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - discard:
              - foo/*.txt
              - bar/{{DEB_HOST_MULTIARCH}}/*.so*
              - baz/script.sh
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_not_installed_file,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "not-installed",
        dh_config_content,
        expected_manifest_content,
    )


def test_migrate_links_files(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        usr/share/target usr/bin/symlink
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            packages:
              foo:
                transformations:
                - create-symlink:
                    path: usr/bin/symlink
                    target: /usr/share/target
    """
    )
    _verify_migrator_generates_parsable_manifest(
        migrate_links_files,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "links",
        dh_config_content,
        expected_manifest_content,
    )


def test_detect_obsolete_substvars(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_obsolete_substvars

    dctrl_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-foo,

    Package: foo
    Architecture: any
    Description: ...
    Depends: ${misc:Depends},
      ${shlibs:Depends},
      bar (>= 1.0~) | baz, ${so:Depends},
    """
    )
    dctrl_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control",
                content=dctrl_content,
            ),
        ]
    )

    migration = run_migrator(
        migrator,
        dctrl_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    msg = (
        "The following relationship substitution variables can be removed from foo:"
        " ${misc:Depends}, ${shlibs:Depends}, ${so:Depends} (Note: https://bugs.debian.org/1067653)"
    )
    assert migration.anything_to_do
    assert migration.warnings == [msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins


def test_detect_obsolete_substvars_remove_field(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_obsolete_substvars

    dctrl_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-foo,

    Package: foo
    Architecture: any
    Description: ...
    Pre-Depends: ${misc:Pre-Depends}
    Depends: bar (>= 1.0~) | baz
    """
    )
    dctrl_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control",
                content=dctrl_content,
            ),
        ]
    )

    migration = run_migrator(
        migrator,
        dctrl_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    msg = (
        "The following relationship fields can be removed from foo: Pre-Depends."
        "  (The content in them would be applied automatically. Note: https://bugs.debian.org/1067653)"
    )
    assert migration.anything_to_do
    assert migration.warnings == [msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins


def test_detect_obsolete_substvars_remove_field_essential(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_obsolete_substvars

    dctrl_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-foo,

    Package: foo
    Architecture: any
    Description: ...
    Essential: yes
    # Obsolete because the package is essential
    Pre-Depends: ${shlibs:Depends}
    Depends: bar (>= 1.0~) | baz
    """
    )
    dctrl_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control",
                content=dctrl_content,
            ),
        ]
    )

    migration = run_migrator(
        migrator,
        dctrl_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    msg = (
        "The following relationship fields can be removed from foo: Pre-Depends."
        "  (The content in them would be applied automatically. Note: https://bugs.debian.org/1067653)"
    )
    assert migration.anything_to_do
    assert migration.warnings == [msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins


def test_detect_obsolete_substvars_remove_field_non_essential(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_obsolete_substvars

    dctrl_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-foo,

    Package: foo
    Architecture: any
    Description: ...
    # This is not obsolete since the package is not essential
    Pre-Depends: ${shlibs:Depends}
    Depends: bar (>= 1.0~) | baz
    """
    )
    dctrl_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control",
                content=dctrl_content,
            ),
        ]
    )

    migration = run_migrator(
        migrator,
        dctrl_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    assert not migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins


def test_detect_dh_addons(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
    accept_any_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_dh_addons_with_zz_integration
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])

    dctrl_no_addons_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13)

    Package: foo
    Architecture: all
    Description: ...
    """
    )

    dctrl_w_addons_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-foo,

    Package: foo
    Architecture: all
    Description: ...
    """
    )

    dctrl_w_migrateable_addons_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-debputy,
                   dh-sequence-numpy3,

    Package: foo
    Architecture: all
    Description: ...
    """
    )

    dctrl_no_addons_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control-without-addons",
                content=dctrl_no_addons_content,
            ),
        ]
    )
    dctrl_w_addons_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control-with-addons",
                content=dctrl_w_addons_content,
            ),
        ]
    )
    dctrl_w_migrateable_addons_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control-with-migrateable-addons",
                content=dctrl_w_migrateable_addons_content,
            ),
        ]
    )
    no_ctrl_msg = (
        "Cannot find debian/control. Detection of unsupported/missing dh-sequence addon"
        " could not be performed. Please ensure the package will Build-Depend on dh-sequence-zz-debputy"
        " and not rely on any other debhelper sequence addons except those debputy explicitly supports."
    )
    missing_debputy_bd_msg = "Missing Build-Depends on dh-sequence-zz-debputy"
    unsupported_sequence_msg = (
        'The dh addon "foo" is not known to work with dh-debputy and might malfunction'
    )

    migration = run_migrator(
        migrator,
        empty_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    assert migration.anything_to_do
    assert migration.warnings == [no_ctrl_msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    migration = run_migrator(
        migrator,
        dctrl_no_addons_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    assert migration.anything_to_do
    assert migration.warnings == [missing_debputy_bd_msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    _assert_unsupported_feature(
        migrator,
        dctrl_w_addons_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    migration = run_migrator(
        migrator,
        dctrl_w_addons_fs,
        empty_manifest_pkg_foo,
        accept_any_migration_issues,
    )

    assert migration.anything_to_do
    assert migration.warnings == [unsupported_sequence_msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    migration = run_migrator(
        migrator,
        dctrl_w_migrateable_addons_fs,
        empty_manifest_pkg_foo,
        accept_any_migration_issues,
    )
    assert not migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert migration.required_plugins == ["numpy3"]


def test_detect_dh_addons_rrr(
    empty_manifest_pkg_foo: HighLevelManifest,
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    migrator = detect_dh_addons_zz_debputy_rrr
    empty_fs = build_virtual_file_system([DEBIAN_DIR_ENTRY])

    dctrl_no_addons_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13)

    Package: foo
    Architecture: all
    Description: ...
    """
    )

    dctrl_w_addons_content = textwrap.dedent(
        """\
    Source: foo
    Build-Depends: debhelper-compat (= 13),
                   dh-sequence-zz-debputy-rrr,
                   dh-sequence-foo,

    Package: foo
    Architecture: all
    Description: ...
    """
    )

    dctrl_no_addons_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control-without-addons",
                content=dctrl_no_addons_content,
            ),
        ]
    )
    dctrl_w_addons_fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                "control",
                fs_path="/nowhere/debian/control-with-addons",
                content=dctrl_w_addons_content,
            ),
        ]
    )
    no_ctrl_msg = (
        "Cannot find debian/control. Detection of unsupported/missing dh-sequence addon"
        " could not be performed. Please ensure the package will Build-Depend on dh-sequence-zz-debputy-rrr."
    )
    missing_debputy_bd_msg = "Missing Build-Depends on dh-sequence-zz-debputy-rrr"

    migration = run_migrator(
        migrator,
        empty_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
        migration_target=INTEGRATION_MODE_DH_DEBPUTY_RRR,
    )
    assert migration.anything_to_do
    assert migration.warnings == [no_ctrl_msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    migration = run_migrator(
        migrator,
        dctrl_no_addons_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )
    assert migration.anything_to_do
    assert migration.warnings == [missing_debputy_bd_msg]
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins

    migration = run_migrator(
        migrator,
        dctrl_w_addons_fs,
        empty_manifest_pkg_foo,
        accept_no_migration_issues,
    )

    assert not migration.anything_to_do
    assert not migration.warnings
    assert not migration.remove_paths_on_success
    assert not migration.rename_paths_on_success
    assert not migration.required_plugins


def test_migrate_bash_completion_file_no_changes(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        compgen -A
    """
    )
    dh_config_name = "bash-completion"
    fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                dh_config_name,
                fs_path=f"/nowhere/debian/{dh_config_name}",
                content=dh_config_content,
            ),
        ]
    )
    migration = run_migrator(
        migrate_bash_completion,
        fs,
        manifest_parser_pkg_foo_factory().build_manifest(),
        accept_no_migration_issues,
    )
    assert not migration.rename_paths_on_success
    assert not migration.remove_paths_on_success
    assert not migration.warnings
    assert not migration.required_plugins


def test_migrate_bash_completion_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
        bar baz
        debian/bar-completion bar
        debian/foo-completion foo
        debian/*.cmpl
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - foo/*
                - debian/*.cmpl
                dest-dir: '{{path:BASH_COMPLETION_DIR}}'
            - install:
                source: bar
                as: '{{path:BASH_COMPLETION_DIR}}/baz'
    """
    )
    expected_renames = [
        ("debian/bar-completion", "debian/foo.bar.bash-completion"),
        ("debian/foo-completion", "debian/foo.bash-completion"),
    ]
    _verify_migrator_generates_parsable_manifest(
        migrate_bash_completion,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "bash-completion",
        dh_config_content,
        expected_manifest_content,
        expected_renamed_paths=expected_renames,
    )


def test_migrate_fish_completions_file_completion_file(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        # I have no idea how a fish-completion file looks.
        something something we win
    """
    )
    dh_config_name = "fish-completions"
    fs = build_virtual_file_system(
        [
            DEBIAN_DIR_ENTRY,
            virtual_path_def(
                dh_config_name,
                fs_path=f"/nowhere/debian/{dh_config_name}",
                content=dh_config_content,
            ),
        ]
    )
    migration = run_migrator(
        migrate_shell_completions,
        fs,
        manifest_parser_pkg_foo_factory().build_manifest(),
        accept_no_migration_issues,
    )
    assert migration.rename_paths_on_success == [
        ("/nowhere/debian/fish-completions", "debian/foo.fish-completion"),
    ]
    assert not migration.remove_paths_on_success
    assert not migration.warnings
    assert not migration.required_plugins


def test_migrate_zsh_completions_file_list(
    manifest_parser_pkg_foo_factory: Callable[[], YAMLManifestParser],
    accept_no_migration_issues: AcceptableMigrationIssues,
) -> None:
    dh_config_content = textwrap.dedent(
        """\
        foo/*
        bar baz
        debian/bar-completion bar
        debian/foo-completion foo
        debian/*.cmpl
    """
    )
    expected_manifest_content = textwrap.dedent(
        """\
            manifest-version: '0.1'
            installations:
            - install:
                sources:
                - foo/*
                - debian/*.cmpl
                dest-dir: '{{path:ZSH_COMPLETION_DIR}}'
            - install:
                source: bar
                as: '{{path:ZSH_COMPLETION_DIR}}/baz'
    """
    )
    expected_renames = [
        ("debian/bar-completion", "debian/foo.bar.zsh-completion"),
        ("debian/foo-completion", "debian/foo.zsh-completion"),
    ]
    _verify_migrator_generates_parsable_manifest(
        migrate_shell_completions,
        manifest_parser_pkg_foo_factory,
        accept_no_migration_issues,
        "zsh-completions",
        dh_config_content,
        expected_manifest_content,
        expected_renamed_paths=expected_renames,
    )



import json
import os
import subprocess
from typing import Optional, Sequence, List, Tuple

from debputy import DEBPUTY_ROOT_DIR
from debputy.commands.debputy_cmd.context import CommandContext
from debputy.deb_packaging_support import setup_control_files
from debputy.filesystem_scan import FSRootDir
from debputy.highlevel_manifest import HighLevelManifest
from debputy.intermediate_manifest import IntermediateManifest
from debputy.plugin.api.impl_types import PackageDataTable
from debputy.util import (
    escape_shell,
    _error,
    compute_output_filename,
    scratch_dir,
    ensure_dir,
    assume_not_none,
    _info,
)

_RRR_DEB_ASSEMBLY_KEYWORD = "debputy/deb-assembly"
_NOTIFIED_ABOUT_FALLBACK_ASSEMBLY = False


def _serialize_intermediate_manifest(members: IntermediateManifest) -> str:
    serial_format = [m.to_manifest() for m in members]
    return json.dumps(serial_format)


def determine_assembly_method(
    package: str,
    intermediate_manifest: IntermediateManifest,
) -> Tuple[bool, bool, List[str]]:
    paths_needing_root = (
        tm for tm in intermediate_manifest if tm.owner != "root" or tm.group != "root"
    )
    matched_path = next(paths_needing_root, None)
    if matched_path is None:
        return False, False, []
    rrr = os.environ.get("DEB_RULES_REQUIRES_ROOT")
    if rrr and _RRR_DEB_ASSEMBLY_KEYWORD in rrr:
        gain_root_cmd = os.environ.get("DEB_GAIN_ROOT_CMD")
        if not gain_root_cmd:
            _error(
                "DEB_RULES_REQUIRES_ROOT contains a debputy keyword but DEB_GAIN_ROOT_CMD does not contain a "
                '"gain root" command'
            )
        return True, False, gain_root_cmd.split()
    if rrr == "no":
        global _NOTIFIED_ABOUT_FALLBACK_ASSEMBLY
        if not _NOTIFIED_ABOUT_FALLBACK_ASSEMBLY:
            _info(
                'Using internal assembly method due to "Rules-Requires-Root" being "no" and dpkg-deb assembly would'
                " require (fake)root for binary packages that needs it."
            )
            _NOTIFIED_ABOUT_FALLBACK_ASSEMBLY = True
        return True, True, []

    _error(
        f'Due to the path "{matched_path.member_path}" in {package}, the package assembly will require (fake)root.'
        " However, this command is not run as root nor was debputy requested to use a root command via"
        f' "Rules-Requires-Root".  Please consider adding "{_RRR_DEB_ASSEMBLY_KEYWORD}" to "Rules-Requires-Root"'
        " in debian/control. Though, due to #1036865, you may have to revert to"
        ' "Rules-Requires-Root: binary-targets" depending on which version of dpkg you need to support.'
        ' Alternatively, you can set "Rules-Requires-Root: no" in debian/control and debputy will assemble'
        " the package anyway. In this case, dpkg-deb will not be used, but the output should be bit-for-bit"
        " compatible with what debputy would have produced with dpkg-deb (and root/fakeroot)."
    )


def assemble_debs(
    context: CommandContext,
    manifest: HighLevelManifest,
    package_data_table: PackageDataTable,
    is_dh_rrr_only_mode: bool,
    *,
    debug_materialization: bool = False,
) -> None:
    parsed_args = context.parsed_args
    output_path = parsed_args.output
    upstream_args = parsed_args.upstream_args
    deb_materialize = str(DEBPUTY_ROOT_DIR / "deb_materialization.py")
    mtime = context.mtime

    for dctrl_bin in manifest.active_packages:
        package = dctrl_bin.name
        dbgsym_package_name = f"{package}-dbgsym"
        dctrl_data = package_data_table[package]
        fs_root = dctrl_data.fs_root
        control_output_fs_path = dctrl_data.control_output_dir.fs_path
        package_metadata_context = dctrl_data.package_metadata_context
        if (
            dbgsym_package_name in package_data_table
            or "noautodbgsym" in manifest.deb_options_and_profiles.deb_build_options
            or "noddebs" in manifest.deb_options_and_profiles.deb_build_options
        ):
            # Discard the dbgsym part if it conflicts with a real package, or
            # we were asked not to build it.
            dctrl_data.dbgsym_info.dbgsym_fs_root = FSRootDir()
            dctrl_data.dbgsym_info.dbgsym_ids.clear()
        dbgsym_fs_root = dctrl_data.dbgsym_info.dbgsym_fs_root
        dbgsym_ids = dctrl_data.dbgsym_info.dbgsym_ids
        intermediate_manifest = manifest.finalize_data_tar_contents(
            package, fs_root, mtime
        )

        setup_control_files(
            dctrl_data,
            manifest,
            dbgsym_fs_root,
            dbgsym_ids,
            package_metadata_context,
            allow_ctrl_file_management=not is_dh_rrr_only_mode,
        )

        needs_root, use_fallback_assembly, gain_root_cmd = determine_assembly_method(
            package, intermediate_manifest
        )

        if not dctrl_bin.is_udeb and any(
            f for f in dbgsym_fs_root.all_paths() if f.is_file
        ):
            # We never built udebs due to #797391. We currently do not generate a control
            # file for it either for the same reason.
            dbgsym_root = dctrl_data.dbgsym_info.dbgsym_root_dir
            if not os.path.isdir(output_path):
                _error(
                    "Cannot produce a dbgsym package when output path is not a directory."
                )
            dbgsym_intermediate_manifest = manifest.finalize_data_tar_contents(
                dbgsym_package_name,
                dbgsym_fs_root,
                mtime,
            )
            _assemble_deb(
                dbgsym_package_name,
                deb_materialize,
                dbgsym_intermediate_manifest,
                mtime,
                os.path.join(dbgsym_root, "DEBIAN"),
                output_path,
                upstream_args,
                is_udeb=dctrl_bin.is_udeb,  # Review this if we ever do dbgsyms for udebs
                use_fallback_assembly=False,
                needs_root=False,
                debug_materialization=debug_materialization,
            )

        _assemble_deb(
            package,
            deb_materialize,
            intermediate_manifest,
            mtime,
            control_output_fs_path,
            output_path,
            upstream_args,
            is_udeb=dctrl_bin.is_udeb,
            use_fallback_assembly=use_fallback_assembly,
            needs_root=needs_root,
            gain_root_cmd=gain_root_cmd,
            debug_materialization=debug_materialization,
        )


def _assemble_deb(
    package: str,
    deb_materialize_cmd: str,
    intermediate_manifest: IntermediateManifest,
    mtime: int,
    control_output_fs_path: str,
    output_path: str,
    upstream_args: Optional[List[str]],
    is_udeb: bool = False,
    use_fallback_assembly: bool = False,
    needs_root: bool = False,
    gain_root_cmd: Optional[Sequence[str]] = None,
    *,
    debug_materialization: bool = False,
) -> None:
    scratch_root_dir = scratch_dir()
    materialization_dir = os.path.join(
        scratch_root_dir, "materialization-dirs", package
    )
    ensure_dir(os.path.dirname(materialization_dir))
    materialize_cmd: List[str] = []
    assert not use_fallback_assembly or not gain_root_cmd
    if needs_root and gain_root_cmd:
        # Only use the gain_root_cmd if we absolutely need it.
        # Note that gain_root_cmd will be empty unless R³ is set to the relevant keyword
        # that would make us use targeted promotion. Therefore, we do not need to check other
        # conditions than the package needing root. (R³: binary-targets implies `needs_root=True`
        # without a gain_root_cmd)
        materialize_cmd.extend(gain_root_cmd)
    materialize_cmd.append(deb_materialize_cmd)
    if debug_materialization:
        materialize_cmd.append("--verbose")
    materialize_cmd.extend(
        [
            "materialize-deb",
            "--intermediate-package-manifest",
            "-",
            "--may-move-control-files",
            "--may-move-data-files",
            "--source-date-epoch",
            str(mtime),
            "--discard-existing-output",
            control_output_fs_path,
            materialization_dir,
        ]
    )
    output = output_path
    if is_udeb:
        materialize_cmd.append("--udeb")
        output = os.path.join(
            output_path, compute_output_filename(control_output_fs_path, True)
        )

    assembly_method = "debputy" if needs_root and use_fallback_assembly else "dpkg-deb"
    combined_materialization_and_assembly = not needs_root
    if combined_materialization_and_assembly:
        materialize_cmd.extend(
            ["--build-method", assembly_method, "--assembled-deb-output", output]
        )

    if upstream_args:
        materialize_cmd.append("--")
        materialize_cmd.extend(upstream_args)

    if combined_materialization_and_assembly:
        _info(
            f"Materializing and assembling {package} via: {escape_shell(*materialize_cmd)}"
        )
    else:
        _info(f"Materializing {package} via: {escape_shell(*materialize_cmd)}")
    proc = subprocess.Popen(materialize_cmd, stdin=subprocess.PIPE)
    proc.communicate(
        _serialize_intermediate_manifest(intermediate_manifest).encode("utf-8")
    )
    if proc.returncode != 0:
        _error(f"{escape_shell(deb_materialize_cmd)} exited with a non-zero exit code!")

    if not combined_materialization_and_assembly:
        build_materialization = [
            deb_materialize_cmd,
            "build-materialized-deb",
            materialization_dir,
            assembly_method,
            "--output",
            output,
        ]
        _info(f"Assembling {package} via: {escape_shell(*build_materialization)}")
        try:
            subprocess.check_call(build_materialization)
        except subprocess.CalledProcessError as e:
            exit_code = f" with exit code {e.returncode}" if e.returncode else ""
            _error(
                f"Assembly command for {package} failed{exit_code}. Please review the output of the command"
                f" for more details on the problem."
            )



import dataclasses
from typing import (
    List,
    NotRequired,
    Union,
    Type,
    Any,
    Mapping,
    Self,
    Tuple,
    Iterable,
    TypeVar,
)

from debputy.manifest_parser.declarative_parser import ParserGenerator
from debputy.manifest_parser.tagging_types import DebputyParsedContent
from debputy.plugin.api.spec import ParserDocumentation

_DEBPUTY_DOC_PARSER_GENERATOR = ParserGenerator()


class AttributeRefDoc(DebputyParsedContent):
    attr: NotRequired[Union[str, List[str]]]
    description: str


class ParserRefDocumentation(DebputyParsedContent):
    title: str
    description: NotRequired[str]
    attributes: NotRequired[List[AttributeRefDoc]]
    undocumented_attributes: NotRequired[List[str]]
    non_mapping_description: NotRequired[str]
    ref_doc_url: NotRequired[str]
    synopsis: NotRequired[str]


class PMRData(ParserRefDocumentation):
    rule_reference: str
    documentation: NotRequired[ParserRefDocumentation]


class DebputyDocReferenceData(DebputyParsedContent):
    pluggable_manifest_rules: NotRequired[List[PMRData]]


DEBPUTY_DOC_REFERENCE_DATA_PARSER = _DEBPUTY_DOC_PARSER_GENERATOR.generate_parser(
    DebputyDocReferenceData
)


D = TypeVar("D")


def _unique(
    topic: str, key_attr_name: str, kd: Iterable[Tuple[str, D]]
) -> Mapping[str, D]:
    seen = set()
    r = {}
    for k, d in kd:
        if k in seen:
            raise ValueError(
                f"The multiple instances of the {key_attr_name} attribute in {topic}"
                f' had the value: "{k}". The value must be unique'
            )
        seen.add(k)
        r[k] = d

    return r


@dataclasses.dataclass(slots=True, frozen=True)
class DebputyParsedDoc:
    pluggable_manifest_rules: Mapping[str, ParserDocumentation]

    @classmethod
    def from_ref_data(cls, docs: DebputyDocReferenceData) -> Self:
        return cls(
            pluggable_manifest_rules=_unique(
                "pluggable-manifest-rules",
                "rule-reference",
                (
                    (d["rule_reference"], ParserDocumentation.from_ref_doc(d))
                    for d in docs.get("pluggable_manifest_rules", [])
                ),
            )
        )


def parser_type_name(v: Union[str, Type[Any]]) -> str:
    if isinstance(v, str):
        return v if v != "<ROOT>" else ""
    return v.__name__



from ...exceptions import (
    DebputyPluginRuntimeError,
    DebputyMetadataAccessError,
)
from .spec import (
    DebputyPluginInitializer,
    PackageProcessingContext,
    MetadataAutoDetector,
    DpkgTriggerType,
    Maintscript,
    VirtualPath,
    BinaryCtrlAccessor,
    PluginInitializationEntryPoint,
    undocumented_attr,
    documented_attr,
    reference_documentation,
    virtual_path_def,
    packager_provided_file_reference_documentation,
)

__all__ = [
    "DebputyPluginInitializer",
    "PackageProcessingContext",
    "MetadataAutoDetector",
    "DpkgTriggerType",
    "Maintscript",
    "BinaryCtrlAccessor",
    "VirtualPath",
    "PluginInitializationEntryPoint",
    "documented_attr",
    "undocumented_attr",
    "reference_documentation",
    "virtual_path_def",
    "DebputyPluginRuntimeError",
    "DebputyMetadataAccessError",
    "packager_provided_file_reference_documentation",
]



import contextlib
import dataclasses
import os
import tempfile
import textwrap
from typing import (
    Iterable,
    Optional,
    Callable,
    Literal,
    Union,
    Iterator,
    overload,
    FrozenSet,
    Sequence,
    TypeVar,
    Any,
    TYPE_CHECKING,
    TextIO,
    BinaryIO,
    Generic,
    ContextManager,
    List,
    Type,
    Tuple,
    get_args,
    Container,
    final,
)

from debian.substvars import Substvars

from debputy import util
from debputy.exceptions import TestPathWithNonExistentFSPathError, PureVirtualPathError
from debputy.interpreter import Interpreter, extract_shebang_interpreter_from_file
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import parse_symbolic_mode
from debputy.packages import BinaryPackage
from debputy.types import S

if TYPE_CHECKING:
    from debputy.manifest_parser.base_types import (
        StaticFileSystemOwner,
        StaticFileSystemGroup,
    )
    from debputy.plugin.api.doc_parsing import ParserRefDocumentation
    from debputy.plugin.api.impl_types import DIPHandler
    from debputy.plugin.debputy.to_be_api_types import BuildSystemRule


DP = TypeVar("DP", bound=DebputyDispatchableType)


PluginInitializationEntryPoint = Callable[["DebputyPluginInitializer"], None]
MetadataAutoDetector = Callable[
    ["VirtualPath", "BinaryCtrlAccessor", "PackageProcessingContext"], None
]
PackageProcessor = Callable[["VirtualPath", None, "PackageProcessingContext"], None]
DpkgTriggerType = Literal[
    "activate",
    "activate-await",
    "activate-noawait",
    "interest",
    "interest-await",
    "interest-noawait",
]
Maintscript = Literal["postinst", "preinst", "prerm", "postrm"]
PackageTypeSelector = Union[Literal["deb", "udeb"], Iterable[Literal["deb", "udeb"]]]
ServiceUpgradeRule = Literal[
    "do-nothing",
    "reload",
    "restart",
    "stop-then-start",
]

DSD = TypeVar("DSD")
ServiceDetector = Callable[
    ["VirtualPath", "ServiceRegistry[DSD]", "PackageProcessingContext"],
    None,
]
ServiceIntegrator = Callable[
    [
        Sequence["ServiceDefinition[DSD]"],
        "BinaryCtrlAccessor",
        "PackageProcessingContext",
    ],
    None,
]

PMT = TypeVar("PMT")
DebputyIntegrationMode = Literal[
    "full",
    "dh-sequence-zz-debputy",
    "dh-sequence-zz-debputy-rrr",
]

INTEGRATION_MODE_FULL: DebputyIntegrationMode = "full"
INTEGRATION_MODE_DH_DEBPUTY_RRR: DebputyIntegrationMode = "dh-sequence-zz-debputy-rrr"
INTEGRATION_MODE_DH_DEBPUTY: DebputyIntegrationMode = "dh-sequence-zz-debputy"
ALL_DEBPUTY_INTEGRATION_MODES: FrozenSet[DebputyIntegrationMode] = frozenset(
    get_args(DebputyIntegrationMode)
)

_DEBPUTY_DISPATCH_METADATA_ATTR_NAME = "_debputy_dispatch_metadata"


def only_integrations(
    *integrations: DebputyIntegrationMode,
) -> Container[DebputyIntegrationMode]:
    return frozenset(integrations)


def not_integrations(
    *integrations: DebputyIntegrationMode,
) -> Container[DebputyIntegrationMode]:
    return ALL_DEBPUTY_INTEGRATION_MODES - frozenset(integrations)


@dataclasses.dataclass(slots=True, frozen=True)
class PackagerProvidedFileReferenceDocumentation:
    description: Optional[str] = None
    format_documentation_uris: Sequence[str] = tuple()

    def replace(self, **changes: Any) -> "PackagerProvidedFileReferenceDocumentation":
        return dataclasses.replace(self, **changes)


def packager_provided_file_reference_documentation(
    *,
    description: Optional[str] = None,
    format_documentation_uris: Optional[Sequence[str]] = tuple(),
) -> PackagerProvidedFileReferenceDocumentation:
    """Provide documentation for a given packager provided file.

    :param description: Textual description presented to the user.
    :param format_documentation_uris: A sequence of URIs to documentation that describes
      the format of the file. Most relevant first.
    :return:
    """
    uris = tuple(format_documentation_uris) if format_documentation_uris else tuple()
    return PackagerProvidedFileReferenceDocumentation(
        description=description,
        format_documentation_uris=uris,
    )


class PathMetadataReference(Generic[PMT]):
    """An accessor to plugin provided metadata

    This is a *short-lived* reference to a piece of metadata.  It should *not* be stored beyond
    the boundaries of the current plugin execution context as it can be become invalid (as an
    example, if the path associated with this path is removed, then this reference become invalid)
    """

    @property
    def is_present(self) -> bool:
        """Determine whether the value has been set

        If the current plugin cannot access the value, then this method unconditionally returns
        `False` regardless of whether the value is there.

        :return: `True` if the value has been set to a not None value (and not been deleted).
          Otherwise, this property is `False`.
        """
        raise NotImplementedError

    @property
    def can_read(self) -> bool:
        """Test whether it is possible to read the metadata

        Note: That the metadata being readable does *not* imply that the metadata is present.

        :return: True if it is possible to read the metadata. This is always True for the
          owning plugin.
        """
        raise NotImplementedError

    @property
    def can_write(self) -> bool:
        """Test whether it is possible to update the metadata

        :return: True if it is possible to update the metadata.
        """
        raise NotImplementedError

    @property
    def value(self) -> Optional[PMT]:
        """Fetch the currently stored value if present.

        :return: The value previously stored if any. Returns `None` if the value was never
          stored, explicitly set to `None` or was deleted.
        """
        raise NotImplementedError

    @value.setter
    def value(self, value: Optional[PMT]) -> None:
        """Replace any current value with the provided value

        This operation is only possible if the path is writable *and* the caller is from
        the owning plugin OR the owning plugin made the reference read-write.
        """
        raise NotImplementedError

    @value.deleter
    def value(self) -> None:
        """Delete any current value.

        This has the same effect as setting the value to `None`.  It has the same restrictions
        as the value setter.
        """
        self.value = None


@dataclasses.dataclass(slots=True)
class PathDef:
    path_name: str
    mode: Optional[int] = None
    mtime: Optional[int] = None
    has_fs_path: Optional[bool] = None
    fs_path: Optional[str] = None
    link_target: Optional[str] = None
    content: Optional[str] = None
    materialized_content: Optional[str] = None


@dataclasses.dataclass(slots=True, frozen=True)
class DispatchablePluggableManifestRuleMetadata(Generic[DP]):
    """NOT PUBLIC API (used internally by part of the public API)"""

    manifest_keywords: Sequence[str]
    dispatched_type: Type[DP]
    unwrapped_constructor: "DIPHandler"
    expected_debputy_integration_mode: Optional[Container[DebputyIntegrationMode]] = (
        None
    )
    online_reference_documentation: Optional["ParserDocumentation"] = None
    apply_standard_attribute_documentation: bool = False
    source_format: Optional[Any] = None


@dataclasses.dataclass(slots=True, frozen=True)
class BuildSystemManifestRuleMetadata(DispatchablePluggableManifestRuleMetadata):
    build_system_impl: Optional[Type["BuildSystemRule"]] = None
    auto_detection_shadow_build_systems: FrozenSet[str] = frozenset()


def virtual_path_def(
    path_name: str,
    /,
    mode: Optional[int] = None,
    mtime: Optional[int] = None,
    fs_path: Optional[str] = None,
    link_target: Optional[str] = None,
    content: Optional[str] = None,
    materialized_content: Optional[str] = None,
) -> PathDef:
    """Define a virtual path for use with examples or, in tests, `build_virtual_file_system`

    :param path_name: The full path. Must start with "./".  If it ends with "/", the path will be interpreted
      as a directory (the `is_dir` attribute will be True).  Otherwise, it will be a symlink or file depending
      on whether a `link_target` is provided.
    :param mode: The mode to use for this path.  Defaults to 0644 for files and 0755 for directories. The mode
      should be None for symlinks.
    :param mtime: Define the last modified time for this path. If not provided, debputy will provide a default
      if the mtime attribute is accessed.
    :param fs_path: Define a file system path for this path.  This causes `has_fs_path` to return True and the
      `fs_path` attribute will return this value.  The test is required to make this path available to the extent
      required. Note that the virtual file system will *not* examine the provided path in any way nor attempt
      to resolve defaults from the path.
    :param link_target: A target for the symlink. Providing a not None value for this parameter will make the
      path a symlink.
    :param content: The content of the path (if opened).  The path must be a file.
    :param materialized_content: Same as `content` except `debputy` will put the contents into a physical file
      as needed. Cannot be used with `content` or `fs_path`.
    :return: An *opaque* object to be passed to `build_virtual_file_system`. While the exact type is provided
      to aid with typing, the type name and its behavior is not part of the API.
    """

    is_dir = path_name.endswith("/")
    is_symlink = link_target is not None

    if is_symlink:
        if mode is not None:
            raise ValueError(
                f'Please do not provide mode for symlinks. Triggered by "{path_name}"'
            )
        if is_dir:
            raise ValueError(
                "Path name looks like a directory, but a symlink target was also provided."
                f' Please remove the trailing slash OR the symlink_target. Triggered by "{path_name}"'
            )

    if content and (is_dir or is_symlink):
        raise ValueError(
            "Content was defined however, the path appears to be a directory a or a symlink"
            f' Please remove the content, the trailing slash OR the symlink_target. Triggered by "{path_name}"'
        )

    if materialized_content is not None:
        if content is not None:
            raise ValueError(
                "The materialized_content keyword is mutually exclusive with the content keyword."
                f' Triggered by "{path_name}"'
            )
        if fs_path is not None:
            raise ValueError(
                "The materialized_content keyword is mutually exclusive with the fs_path keyword."
                f' Triggered by "{path_name}"'
            )
    return PathDef(
        path_name,
        mode=mode,
        mtime=mtime,
        has_fs_path=bool(fs_path) or materialized_content is not None,
        fs_path=fs_path,
        link_target=link_target,
        content=content,
        materialized_content=materialized_content,
    )


class PackageProcessingContext:
    """Context for auto-detectors of metadata and package processors (no instantiation)

    This object holds some context related data for the metadata detector or/and package
    processors.  It may receive new attributes in the future.
    """

    __slots__ = ()

    @property
    def binary_package(self) -> BinaryPackage:
        """The binary package stanza from `debian/control`"""
        raise NotImplementedError

    @property
    def binary_package_version(self) -> str:
        """The version of the binary package

        Note this never includes the binNMU version for arch:all packages, but it may for arch:any.
        """
        raise NotImplementedError

    @property
    def related_udeb_package(self) -> Optional[BinaryPackage]:
        """An udeb related to this binary package (if any)"""
        raise NotImplementedError

    @property
    def related_udeb_package_version(self) -> Optional[str]:
        """The version of the related udeb package (if present)

        Note this never includes the binNMU version for arch:all packages, but it may for arch:any.
        """
        raise NotImplementedError

    def accessible_package_roots(self) -> Iterable[Tuple[BinaryPackage, "VirtualPath"]]:
        raise NotImplementedError

    # """The source package stanza from `debian/control`"""
    # source_package: SourcePackage


class DebputyPluginInitializer:
    __slots__ = ()

    def packager_provided_file(
        self,
        stem: str,
        installed_path: str,
        *,
        default_mode: int = 0o0644,
        default_priority: Optional[int] = None,
        allow_name_segment: bool = True,
        allow_architecture_segment: bool = False,
        post_formatting_rewrite: Optional[Callable[[str], str]] = None,
        packageless_is_fallback_for_all_packages: bool = False,
        reservation_only: bool = False,
        reference_documentation: Optional[
            PackagerProvidedFileReferenceDocumentation
        ] = None,
    ) -> None:
        """Register a packager provided file (debian/<pkg>.foo)

        Register a packager provided file that debputy should automatically detect and install for the
        packager (example `debian/foo.tmpfiles` -> `debian/foo/usr/lib/tmpfiles.d/foo.conf`).  A packager
        provided file typically identified by a package prefix and a "stem" and by convention placed
        in the `debian/` directory.

        Like debhelper, debputy also supports the `foo.bar.tmpfiles` variant where the file is to be
        installed into the `foo` package but be named after the `bar` segment rather than the package name.
        This feature can be controlled via the `allow_name_segment` parameter.

        :param stem: The "stem" of the file. This would be the `tmpfiles` part of `debian/foo.tmpfiles`.
          Note that this value must be unique across all registered packager provided files.
        :param installed_path: A format string describing where the file should be installed. Would be
          `/usr/lib/tmpfiles.d/{name}.conf` from the example above.

          The caller should provide a string with one or more of the placeholders listed below (usually `{name}`
          should be one of them). The format affect the entire path.

          The following placeholders are supported:
            * `{name}` - The name in the name segment (defaulting the package name if no name segment is given)
            * `{priority}` / `{priority:02}` - The priority of the file. Only provided priorities are used (that
               is, default_priority is not None).  The latter variant ensuring that the priority takes at least
               two characters and the `0` character is left-padded for priorities that takes less than two
               characters.
            * `{owning_package}` - The name of the package.  Should only be used when `{name}` alone is insufficient.
              If you do not want the "name" segment in the first place, use `allow_name_segment=False` instead.

          The path is always interpreted as relative to the binary package root.

        :param default_mode: The mode the installed file should have by default. Common options are 0o0644 (the default)
          or 0o0755 (for files that must be executable).
        :param allow_architecture_segment: If True, the file may have an optional "architecture" segment at the end
           (`foo.tmpfiles.amd64`), which marks it architecture specific. When False, debputy will detect the
           "architecture" segment and report the use as an error.  Note the architecture segment is only allowed for
           arch:any packages. If a file targeting an arch:all package uses an architecture specific file it will
           always result in an error.
        :param allow_name_segment: If True, the file may have an optional "name" segment after the package name prefix.
           (`foo.<name-here>.tmpfiles`). When False, debputy will detect the "name" segment and report the use as an
           error.
        :param default_priority: Special-case option for packager files that are installed into directories that have
          "parse ordering" or "priority".  These files will generally be installed as something like `20-foo.conf`
          where the `20-` denotes their "priority".  If the plugin is registering such a file type, then it should
          provide a default priority.

          The following placeholders are supported:
            * `{name}` - The name in the name segment (defaulting the package name if no name segment is given)
            * `{priority}` - The priority of the file. Only provided priorities are used (that is, default_priority
               is not None)
            * `{owning_package}` - The name of the package.  Should only be used when `{name}` alone is insufficient.
              If you do not want the "name" segment in the first place, use `allow_name_segment=False` instead.
        :param post_formatting_rewrite: An optional "name correcting" callback. It receives the formatted name and can
          do any transformation required. The primary use-case for this is to replace "forbidden" characters. The most
          common case for debputy itself is to replace "." with "_" for tools that refuse to work with files containing
          "." (`lambda x: x.replace(".", "_")`).  The callback operates on basename of formatted version of the
          `installed_path` and the callback should return the basename.
        :param packageless_is_fallback_for_all_packages: If True, the packageless variant (such as, `debian/changelog`)
          is a fallback for every package.
        :param reference_documentation: Reference documentation for the packager provided file. Use the
           packager_provided_file_reference_documentation function to provide the value for this parameter.
        :param reservation_only: When True, tell debputy that the plugin reserves this packager provided file, but that
          debputy should not actually install it automatically.  This is useful in the cases, where the plugin
          needs to process the file before installing it.  The file will be marked as provided by this plugin. This
          enables introspection and detects conflicts if other plugins attempts to claim the file.
        """
        raise NotImplementedError

    def metadata_or_maintscript_detector(
        self,
        auto_detector_id: str,
        auto_detector: MetadataAutoDetector,
        *,
        package_type: PackageTypeSelector = "deb",
    ) -> None:
        """Provide a pre-assembly hook that can affect the metadata/maintscript of binary ("deb") packages

        The provided hook will be run once per binary package to be assembled, and it can see all the content
        ("data.tar") planned to be included in the deb. The hook may do any *read-only* analysis of this content
        and provide metadata, alter substvars or inject maintscript snippets.  However, the hook must *not*
        change the content ("data.tar") part of the deb.

        The hook will be run unconditionally for all binary packages built. When the hook does not apply to all
        packages, it must provide its own (internal) logic for detecting whether it is relevant and reduced itself
        to a no-op if it should not apply to the current package.

        Hooks are run in "some implementation defined order" and should not rely on being run before or after
        any other hook.

        The hooks are only applied to packages defined in `debian/control`. Notably, the metadata detector will
        not apply to auto-generated `-dbgsym` packages (as those are not listed explicitly in `debian/control`).

        :param auto_detector_id: A plugin-wide unique ID for this detector. Packagers may use this ID for disabling
          the detector and accordingly the ID is part of the plugin's API toward the packager.
        :param auto_detector: The code to be called that will be run at the metadata generation state (once for each
          binary package).
        :param package_type: Which kind of packages this metadata detector applies to.  The package type is generally
          defined by `Package-Type` field in the binary package. The default is to only run for regular `deb` packages
          and ignore `udeb` packages.
        """
        raise NotImplementedError

    def manifest_variable(
        self,
        variable_name: str,
        value: str,
        variable_reference_documentation: Optional[str] = None,
    ) -> None:
        """Provide a variable that can be used in the package manifest

            >>> # Enable users to use "{{path:BASH_COMPLETION_DIR}}/foo" in their manifest.
            >>> api.manifest_variable(  # doctest: +SKIP
            ...     "path:BASH_COMPLETION_DIR",
            ...     "/usr/share/bash-completion/completions",
            ...     variable_reference_documentation="Directory to install bash completions into",
            ... )

        :param variable_name: The variable name.
        :param value: The value the variable should resolve to.
        :param variable_reference_documentation: A short snippet of reference documentation that explains
          the purpose of the variable.
        """
        raise NotImplementedError


class MaintscriptAccessor:
    __slots__ = ()

    def on_configure(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
        skip_on_rollback: bool = False,
    ) -> None:
        """Provide a snippet to be run when the package is about to be "configured"

        This condition is the most common "post install" condition and covers the two
        common cases:
          * On initial install, OR
          * On upgrade

        In dpkg maintscript terms, this method roughly corresponds to postinst containing
             `if [ "$1" = configure ]; then <snippet>; fi`

        Additionally, the condition will by default also include rollback/abort scenarios such as "above-remove",
        which is normally what you want but most people forget about.

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param skip_on_rollback: By default, this condition will also cover common rollback scenarios. This
          is normally what you want (or benign in most cases due to the idempotence requirement for maintscripts).
          However, you can disable the rollback cases, leaving only "On initial install OR On upgrade".
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_initial_install(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package is about to be "configured" for the first time

        The snippet will only be run on the first time the package is installed (ever or since last purge).
        Note that "first" does not mean "exactly once" as dpkg does *not* provide such semantics. There are two
        common cases where this can snippet can be run multiple times for the same system (and why the snippet
        must still be idempotent):

          1) The package is installed (1), then purged and then installed again (2).  This can partly be mitigated
             by having an `on_purge` script to do clean up.

          2) As the package is installed, the `postinst` script terminates prematurely (Disk full, power loss, etc.).
             The user resolves the problem and runs `dpkg --configure <pkg>`, which in turn restarts the script
             from the beginning.  This is why scripts must be idempotent in general.

        In dpkg maintscript terms, this method roughly corresponds to postinst containing
             `if [ "$1" = configure ] && [ -z "$2" ]; then <snippet>; fi`

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_upgrade(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package is about to be "configured" after an upgrade

        The snippet will only be run on any upgrade (that is, it will be skipped on the initial install).

        In dpkg maintscript terms, this method roughly corresponds to postinst containing
             `if [ "$1" = configure ] && [ -n "$2" ]; then <snippet>; fi`

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_upgrade_from(
        self,
        version: str,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package is about to be "configured" after an upgrade from a given version

        The snippet will only be run on any upgrade (that is, it will be skipped on the initial install).

        In dpkg maintscript terms, this method roughly corresponds to postinst containing
             `if [ "$1" = configure ] && dpkg --compare-versions le-nl "$2" ; then <snippet>; fi`

        :param version: The version to upgrade from
        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_before_removal(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package is about to be removed

        The snippet will be run before dpkg removes any files.

        In dpkg maintscript terms, this method roughly corresponds to prerm containing
             `if [ "$1" = remove ] ; then <snippet>; fi`

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_removed(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package has been removed

        The snippet will be run after dpkg removes the package content from the file system.

        **WARNING**: The snippet *cannot* rely on dependencies and must rely on `Essential: yes` packages.

        In dpkg maintscript terms, this method roughly corresponds to postrm containing
             `if [ "$1" = remove ] ; then <snippet>; fi`

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def on_purge(
        self,
        run_snippet: str,
        /,
        indent: Optional[bool] = None,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run when the package is being purged.

        The snippet will when the package is purged from the system.

        **WARNING**: The snippet *cannot* rely on dependencies and must rely on `Essential: yes` packages.

        In dpkg maintscript terms, this method roughly corresponds to postrm containing
             `if [ "$1" = purge ] ; then <snippet>; fi`

        :param run_snippet: The actual shell snippet to be run in the given condition.  The snippet must be idempotent.
          The snippet may contain newlines as necessary, which will make the result more readable.  Additionally, the
          snippet may contain '{{FOO}}' substitutions by default.
        :param indent: If True, the provided snippet will be indented to fit the condition provided by debputy.
          In most cases, this is safe to do and provides more readable scripts. However, it may cause issues
          with some special shell syntax (such as "Heredocs"). When False, the snippet will *not* be re-indented.
          You are recommended to do 4 spaces of indentation when indent is False for readability.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def unconditionally_in_script(
        self,
        maintscript: Maintscript,
        run_snippet: str,
        /,
        perform_substitution: bool = True,
    ) -> None:
        """Provide a snippet to be run in a given script

        Run a given snippet unconditionally from a given script.  The snippet must contain its own conditional
        for when it should be run.

        :param maintscript: The maintscript to insert the snippet into.
        :param run_snippet: The actual shell snippet to be run.  The snippet will be run unconditionally and should
          contain its own conditions as necessary. The snippet must be idempotent. The snippet may contain newlines
          as necessary, which will make the result more readable.  Additionally, the snippet may contain '{{FOO}}'
          substitutions by default.
        :param perform_substitution: When True, `{{FOO}}` will be substituted in the snippet. When False, no
          substitution is provided.
        """
        raise NotImplementedError

    def escape_shell_words(self, *args: str) -> str:
        """Provide sh-shell escape of strings

          `assert escape_shell("foo", "fu bar", "baz") == 'foo "fu bar" baz'`

        This is useful for ensuring file names and other "input" are considered one parameter even when they
        contain spaces or shell meta-characters.

        :param args: The string(s) to be escaped.
        :return: Each argument escaped so that each argument becomes a single "word" and then all these words are
          joined by a single space.
        """
        return util.escape_shell(*args)


class BinaryCtrlAccessor:
    __slots__ = ()

    def dpkg_trigger(self, trigger_type: DpkgTriggerType, trigger_target: str) -> None:
        """Register a declarative dpkg level trigger

        The provided trigger will be added to the package's metadata (the triggers file of the control.tar).

        If the trigger has already been added previously, a second call with the same trigger data will be ignored.
        """
        raise NotImplementedError

    @property
    def maintscript(self) -> MaintscriptAccessor:
        """Attribute for manipulating maintscripts"""
        raise NotImplementedError

    @property
    def substvars(self) -> "FlushableSubstvars":
        """Attribute for manipulating dpkg substvars (deb-substvars)"""
        raise NotImplementedError


class VirtualPath:
    __slots__ = ()

    @property
    def name(self) -> str:
        """Basename of the path a.k.a. last segment of the path

        In a path "usr/share/doc/pkg/changelog.gz" the basename is "changelog.gz".

        For a directory, the basename *never* ends with a `/`.
        """
        raise NotImplementedError

    @property
    def iterdir(self) -> Iterable["VirtualPath"]:
        """Returns an iterable that iterates over all children of this path

        For directories, this returns an iterable of all children. For non-directories,
        the iterable is always empty.
        """
        raise NotImplementedError

    def lookup(self, path: str) -> Optional["VirtualPath"]:
        """Perform a path lookup relative to this path

        As an example `doc_dir = fs_root.lookup('./usr/share/doc')`

        If the provided path starts with `/`, then the lookup is performed relative to the
        file system root.  That is, you can assume the following to always be True:

            `fs_root.lookup("usr") == any_path_beneath_fs_root.lookup('/usr')`

        Note: This method requires the path to be attached (see `is_detached`) regardless of
        whether the lookup is relative or absolute.

        If the path traverse a symlink, the symlink will be resolved.

        :param path: The path to look. Can contain "." and ".." segments.  If starting with `/`,
          look up is performed relative to the file system root, otherwise the lookup is relative
          to this path.
        :return: The path object for the desired path if it can be found. Otherwise, None.
        """
        raise NotImplementedError

    def all_paths(self) -> Iterable["VirtualPath"]:
        """Iterate over this path and all of its descendants (if any)

        If used on the root path, then every path in the package is returned.

        The iterable is ordered, so using the order in output will be produce
        bit-for-bit reproducible output. Additionally, a directory will always
        be seen before its descendants. Otherwise, the order is implementation
        defined.

        The iteration is lazy and as a side effect do account for some obvious
        mutation. Like if the current path is removed, then none of its children
        will be returned (provided mutation happens before the lazy iteration
        was required to resolve it). Likewise, mutation of the directory will
        also work (again, provided mutation happens before the lazy iteration order).

        :return: An ordered iterable of this path followed by its descendants.
        """
        raise NotImplementedError

    @property
    def is_detached(self) -> bool:
        """Returns True if this path is detached

        Paths that are detached from the file system will not be present in the package and
        most operations are unsafe on them. This usually only happens if the path or one of
        its parent directories are unlinked (rm'ed) from the file system tree.

        All paths are attached by default and will only become detached as a result of
        an action to mutate the virtual file system.  Note that the file system may not
        always be manipulated.

        :return: True if the entry is detached. Detached entries should be discarded, so they
        can be garbage collected.
        """
        raise NotImplementedError

    # The __getitem__ behaves like __getitem__ from Dict but __iter__ would ideally work like a Sequence.
    # However, that does not feel compatible, so lets force people to use .children instead for the Sequence
    # behavior to avoid surprises for now.
    # (Maybe it is a non-issue, but it is easier to add the API later than to remove it once we have committed
    # to using it)
    __iter__ = None

    def __getitem__(self, key: object) -> "VirtualPath":
        """Lookup a (direct) child by name

        Ignoring the possible `KeyError`, then the following are the same:
            `fs_root["usr"] == fs_root.lookup('usr')`

        Note that unlike `.lookup` this can only locate direct children.
        """
        raise NotImplementedError

    def __delitem__(self, key) -> None:
        """Remove a child from this node if it exists

        If that child is a directory, then the entire tree is removed (like `rm -fr`).
        """
        raise NotImplementedError

    def get(self, key: str) -> "Optional[VirtualPath]":
        """Lookup a (direct) child by name

        The following are the same:
            `fs_root.get("usr") == fs_root.lookup('usr')`

        Note that unlike `.lookup` this can only locate direct children.
        """
        try:
            return self[key]
        except KeyError:
            return None

    def __contains__(self, item: object) -> bool:
        """Determine if this path includes a given child (either by object or string)

        Examples:

            if 'foo' in dir: ...
        """
        if isinstance(item, VirtualPath):
            return item.parent_dir is self
        if not isinstance(item, str):
            return False
        m = self.get(item)
        return m is not None

    @property
    def path(self) -> str:
        """Returns the "full" path for this file system entry

        This is the path that debputy uses to refer to this file system entry. It is always
        normalized. Use the `absolute` attribute for how the path looks
        when the package is installed. Alternatively, there is also `fs_path`, which is the
        path to the underlying file system object (assuming there is one). That is the one
        you need if you want to read the file.

        This is attribute is mostly useful for debugging or for looking up the path relative
        to the "root" of the virtual file system that debputy maintains.

        If the path is detached (see `is_detached`), then this method returns the path as it
        was known prior to being detached.
        """
        raise NotImplementedError

    @property
    def absolute(self) -> str:
        """Returns the absolute version of this path

        This is how to refer to this path when the package is installed.

        If the path is detached (see `is_detached`), then this method returns the last known location
        of installation (prior to being detached).

        :return: The absolute path of this file as it would be on the installed system.
        """
        p = self.path.lstrip(".")
        if not p.startswith("/"):
            return f"/{p}"
        return p

    @property
    def parent_dir(self) -> Optional["VirtualPath"]:
        """The parent directory of this path

        Note this operation requires the path is "attached" (see `is_detached`).  All paths are attached
        by default but unlinking paths will cause them to become detached.

        :return: The parent path or None for the root.
        """
        raise NotImplementedError

    def stat(self) -> os.stat_result:
        """Attempt to do stat of the underlying path (if it exists)

        *Avoid* using `stat()` whenever possible where a more specialized attribute exist.  The
        `stat()` call returns the data from the file system and often, `debputy` does *not* track
        its state in the file system.  As an example, if you want to know the file system mode of
        a path, please use the `mode` attribute instead.

        This never follow symlinks (it behaves like `os.lstat`). It will raise an error
        if the path is not backed by a file system object (that is, `has_fs_path` is False).

        :return: The stat result or an error.
        """
        raise NotImplementedError()

    @property
    def size(self) -> int:
        """Resolve the file size (`st_size`)

        This may be using `stat()` and therefore `fs_path`.

        :return: The size of the file in bytes
        """
        return self.stat().st_size

    @property
    def mode(self) -> int:
        """Determine the mode bits of this path object

        Note that:
         * like with `stat` above, this never follows symlinks.
         * the mode returned by this method is not always a 1:1 with the mode in the
           physical file system. As an optimization, `debputy` skips unnecessary writes
           to the underlying file system in many cases.


        :return: The mode bits for the path.
        """
        raise NotImplementedError

    @mode.setter
    def mode(self, new_mode: int) -> None:
        """Set the octal file mode of this path

        Note that:
         * this operation will fail if `path.is_read_write` returns False.
         * this operation is generally *not* synced to the physical file system (as
           an optimization).

        :param new_mode: The new octal mode for this path.  Note that `debputy` insists
          that all paths have the `user read bit` and, for directories also, the
          `user execute bit`.  The absence of these minimal mode bits causes hard to
          debug errors.
        """
        raise NotImplementedError

    @property
    def is_executable(self) -> bool:
        """Determine whether a path is considered executable

        Generally, this means that at least one executable bit is set. This will
        basically always be true for directories as directories need the execute
        parameter to be traversable.

        :return: True if the path is considered executable with its current mode
        """
        return bool(self.mode & 0o0111)

    def chmod(self, new_mode: Union[int, str]) -> None:
        """Set the file mode of this path

        This is similar to setting the `mode` attribute. However, this method accepts
        a string argument, which will be parsed as a symbolic mode (example: `u+rX,go=rX`).

        Note that:
         * this operation will fail if `path.is_read_write` returns False.
         * this operation is generally *not* synced to the physical file system (as
           an optimization).

        :param new_mode: The new mode for this path.
          Note that `debputy` insists that all paths have the `user read bit` and, for
          directories also, the `user execute bit`.  The absence of these minimal mode
          bits causes hard to debug errors.
        """
        if isinstance(new_mode, str):
            segments = parse_symbolic_mode(new_mode, None)
            final_mode = self.mode
            is_dir = self.is_dir
            for segment in segments:
                final_mode = segment.apply(final_mode, is_dir)
            self.mode = final_mode
        else:
            self.mode = new_mode

    def chown(
        self,
        owner: Optional["StaticFileSystemOwner"],
        group: Optional["StaticFileSystemGroup"],
    ) -> None:
        """Change the owner/group of this path

        :param owner: The desired owner definition for this path. If None, then no change of owner is performed.
        :param group: The desired  group definition for this path. If None, then no change of group is performed.
        """
        raise NotImplementedError

    @property
    def mtime(self) -> float:
        """Determine the mtime of this path object

        Note that:
         * like with `stat` above, this never follows symlinks.
         * the mtime returned has *not* been clamped against ´SOURCE_DATE_EPOCH`. Timestamp
           normalization is handled later by `debputy`.
         * the mtime returned by this method is not always a 1:1 with the mtime in the
           physical file system. As an optimization, `debputy` skips unnecessary writes
           to the underlying file system in many cases.

        :return: The mtime for the path.
        """
        raise NotImplementedError

    @mtime.setter
    def mtime(self, new_mtime: float) -> None:
        """Set the mtime of this path

        Note that:
         * this operation will fail if `path.is_read_write` returns False.
         * this operation is generally *not* synced to the physical file system (as
           an optimization).

        :param new_mtime: The new mtime of this path. Note that the caller does not need to
          account for `SOURCE_DATE_EPOCH`. Timestamp normalization is handled later.
        """
        raise NotImplementedError

    def readlink(self) -> str:
        """Determine the link target of this path assuming it is a symlink

        For paths where `is_symlink` is True, this already returns a link target even when
        `has_fs_path` is False.

        :return: The link target of the path or an error is this is not a symlink
        """
        raise NotImplementedError()

    @overload
    def open(
        self,
        *,
        byte_io: Literal[False] = False,
        buffering: int = -1,
    ) -> TextIO: ...

    @overload
    def open(
        self,
        *,
        byte_io: Literal[True],
        buffering: int = -1,
    ) -> BinaryIO: ...

    @overload
    def open(
        self,
        *,
        byte_io: bool,
        buffering: int = -1,
    ) -> Union[TextIO, BinaryIO]: ...

    def open(
        self,
        *,
        byte_io: bool = False,
        buffering: int = -1,
    ) -> Union[TextIO, BinaryIO]:
        """Open the file for reading.  Usually used with a context manager

        By default, the file is opened in text mode (utf-8). Binary mode can be requested
        via the `byte_io` parameter.  This operation is only valid for files (`is_file` returns
        `True`). Usage on symlinks and directories will raise exceptions.

        This method *often* requires the `fs_path` to be present.  However, tests as a notable
        case can inject content without having the `fs_path` point to a real file. (To be clear,
        such tests are generally expected to ensure `has_fs_path` returns `True`).


        :param byte_io: If True, open the file in binary mode (like `rb` for `open`)
        :param buffering: Same as open(..., buffering=...) where supported. Notably during
          testing, the content may be purely in memory and use a BytesIO/StringIO
          (which does not accept that parameter, but then it is buffered in a different way)
        :return: The file handle.
        """

        if not self.is_file:
            raise TypeError(f"Cannot open {self.path} for reading: It is not a file")

        if byte_io:
            return open(self.fs_path, "rb", buffering=buffering)
        return open(self.fs_path, "rt", encoding="utf-8", buffering=buffering)

    @property
    def fs_path(self) -> str:
        """Request the underling fs_path of this path

        Only available when `has_fs_path` is True.  Generally this should only be used for files to read
        the contents of the file and do some action based on the parsed result.

        The path should only be used for read-only purposes as debputy may assume that it is safe to have
        multiple paths pointing to the same file system path.

        Note that:
          * This is often *not* available for directories and symlinks.
          * The debputy in-memory file system overrules the physical file system. Attempting to "fix" things
            by using `os.chmod` or `os.unlink`'ing files, etc. will generally not do as you expect. Best case,
            your actions are ignored and worst case it will cause the build to fail as it violates debputy's
            internal invariants.

        :return: The path to the underlying file system object on the build system or an error if no such
        file exist (see `has_fs_path`).
        """
        raise NotImplementedError()

    @property
    def is_dir(self) -> bool:
        """Determine if this path is a directory

        Never follows symlinks.

        :return: True if this path is a directory. False otherwise.
        """
        raise NotImplementedError()

    @property
    def is_file(self) -> bool:
        """Determine if this path is a directory

        Never follows symlinks.

        :return: True if this path is a regular file. False otherwise.
        """
        raise NotImplementedError()

    @property
    def is_symlink(self) -> bool:
        """Determine if this path is a symlink

        :return: True if this path is a symlink. False otherwise.
        """
        raise NotImplementedError()

    @property
    def has_fs_path(self) -> bool:
        """Determine whether this path is backed by a file system path

        :return: True if this path is backed by a file system object on the build system.
        """
        raise NotImplementedError()

    @property
    def is_read_write(self) -> bool:
        """When true, the file system entry may be mutated

        Read-write rules are:

        +--------------------------+-------------------+------------------------+
        | File system              | From / Inside     | Read-Only / Read-Write |
        +--------------------------+-------------------+------------------------+
        | Source directory         | Any context       | Read-Only              |
        | Binary staging directory | Package Processor | Read-Write             |
        | Binary staging directory | Metadata Detector | Read-Only              |
        +--------------------------+-------------------+------------------------+

        These rules apply to the virtual file system (`debputy` cannot enforce
        these rules in the underlying file system). The `debputy` code relies
        on these rules for its logic in multiple places to catch bugs and for
        optimizations.

        As an example, the reason why the file system is read-only when Metadata
        Detectors are run is based the contents of the file system has already
        been committed. New files will not be included, removals of existing
        files will trigger a hard error when the package is assembled, etc.
        To avoid people spending hours debugging why their code does not work
        as intended, `debputy` instead throws a hard error if you try to mutate
        the file system when it is read-only mode to "fail fast".

        :return: Whether file system mutations are permitted.
        """
        return False

    def mkdir(self, name: str) -> "VirtualPath":
        """Create a new subdirectory of the current path

        :param name: Basename of the new directory. The directory must not contain a path
          with this basename.
        :return: The new subdirectory
        """
        raise NotImplementedError

    def mkdirs(self, path: str) -> "VirtualPath":
        """Ensure a given path exists and is a directory.

        :param path: Path to the directory to create. Any parent directories will be
          created as needed. If the path already exists and is a directory, then it
          is returned.  If any part of the path exists and that is not a directory,
          then the `mkdirs` call will raise an error.
        :return: The directory denoted by the given path
        """
        raise NotImplementedError

    def add_file(
        self,
        name: str,
        *,
        unlink_if_exists: bool = True,
        use_fs_path_mode: bool = False,
        mode: int = 0o0644,
        mtime: Optional[float] = None,
    ) -> ContextManager["VirtualPath"]:
        """Add a new regular file as a child of this path

        This method will insert a new file into the virtual file system as a child
        of the current path (which must be a directory).  The caller must use the
        return value as a context manager (see example).  During the life-cycle of
        the managed context, the caller can fill out the contents of the file
        from the new path's `fs_path` attribute. The `fs_path` will exist as an
        empty file when the context manager is entered.

        Once the context manager exits, mutation of the `fs_path` is no longer permitted.

          >>> import subprocess
          >>> path = ...                                                                 # doctest: +SKIP
          >>> with path.add_file("foo") as new_file, open(new_file.fs_path, "w") as fd:  # doctest: +SKIP
          ...     fd.writelines(["Some", "Content", "Here"])

        The caller can replace the provided `fs_path` entirely provided at the end result
        (when the context manager exits) is a regular file with no hard links.

        Note that this operation will fail if `path.is_read_write` returns False.

        :param name: Basename of the new file
        :param unlink_if_exists: If the name was already in use, then either an exception is thrown
           (when `unlink_if_exists` is False) or the path will be removed via ´unlink(recursive=False)`
           (when `unlink_if_exists` is True)
        :param use_fs_path_mode: When True, the file created will have this mode in the physical file
          system. When the context manager exists, `debputy` will refresh its mode to match the mode
          in the physical file system.  This is primarily useful if the caller uses a subprocess to
          mutate the path and the file mode is relevant for this tool (either as input or output).
          When the parameter is false, the new file is guaranteed to be readable and writable for
          the current user. However, no other guarantees are given (not even that it matches the
          `mode` parameter and any changes to the mode in the physical file system will be ignored.
        :param mode: This is the initial file mode. Note the `use_fs_path_mode` parameter for how
          this interacts with the physical file system.
        :param mtime: If the caller has a more accurate mtime than the mtime of the generated file,
          then it can be provided here. Note that all mtimes will later be clamped based on
          `SOURCE_DATE_EPOCH`. This parameter is only for when the conceptual mtime of this path
          should be earlier than `SOURCE_DATE_EPOCH`.
        :return: A Context manager that upon entering provides a `VirtualPath` instance for the
                 new file. The instance remains valid after the context manager exits (assuming it exits
                 successfully), but the file denoted by `fs_path` must not be changed after the context
                 manager exits
        """
        raise NotImplementedError

    def replace_fs_path_content(
        self,
        *,
        use_fs_path_mode: bool = False,
    ) -> ContextManager[str]:
        """Replace the contents of this file via inline manipulation

        Used as a context manager to provide the fs path for manipulation.

        Example:
            >>> import subprocess
            >>> path = ...                                       # doctest: +SKIP
            >>> with path.replace_fs_path_content() as fs_path:  # doctest: +SKIP
            ...    subprocess.check_call(['strip', fs_path])     # doctest: +SKIP

        The provided file system path should be manipulated inline. The debputy framework may
        copy it first as necessary and therefore the provided fs_path may be different from
        `path.fs_path` prior to entering the context manager.

        Note that this operation will fail if `path.is_read_write` returns False.

        If the mutation causes the returned `fs_path` to be a non-file or a hard-linked file
        when the context manager exits, `debputy` will raise an error at that point. To preserve
        the internal invariants of `debputy`, the path will be unlinked as `debputy` cannot
        reliably restore the path.

        :param use_fs_path_mode: If True, any changes to the mode on the physical FS path will be
          recorded as the desired mode of the file when the contextmanager ends.  The provided FS path
          with start with the current mode when `use_fs_path_mode` is True. Otherwise, `debputy` will
          ignore the mode of the file system entry and reuse its own current mode
          definition.
        :return: A Context manager that upon entering provides the path to a muable (copy) of
                 this path's `fs_path` attribute. The file on the underlying path may be mutated however
                 the caller wishes until the context manager exits.
        """
        raise NotImplementedError

    def add_symlink(self, link_name: str, link_target: str) -> "VirtualPath":
        """Add a new regular file as a child of this path

        This will create a new symlink inside the current path. If the path already exists,
        the existing path will be unlinked via `unlink(recursive=False)`.

        Note that this operation will fail if `path.is_read_write` returns False.

        :param link_name: The basename of the link file entry.
        :param link_target: The target of the link.  Link target normalization will
          be handled by `debputy`, so the caller can use relative or absolute paths.
          (At the time of writing, symlink target normalization happens late)
        :return: The newly created symlink.
        """
        raise NotImplementedError

    def unlink(self, *, recursive: bool = False) -> None:
        """Unlink a file or a directory

        This operation will remove the path from the file system (causing `is_detached` to return True).

        When the path is a:

         * symlink, then the symlink itself is removed. The target (if present) is not affected.
         * *non-empty* directory, then the `recursive` parameter decides the outcome. An empty
           directory will be removed regardless of the value of `recursive`.

        Note that:
          * the root directory cannot be deleted.
          * this operation will fail if `path.is_read_write` returns False.

        :param recursive: If True, then non-empty directories will be unlinked as well removing everything inside them
          as well.  When False, an error is raised if the path is a non-empty directory
        """
        raise NotImplementedError

    def interpreter(self) -> Optional[Interpreter]:
        """Determine the interpreter of the file (`#!`-line details)

        Note: this method is only applicable for files (`is_file` is True).

        :return: The detected interpreter if present or None if no interpreter can be detected.
        """
        if not self.is_file:
            raise TypeError("Only files can have interpreters")
        try:
            with self.open(byte_io=True, buffering=4096) as fd:
                return extract_shebang_interpreter_from_file(fd)
        except (PureVirtualPathError, TestPathWithNonExistentFSPathError):
            return None

    def metadata(
        self,
        metadata_type: Type[PMT],
    ) -> PathMetadataReference[PMT]:
        """Fetch the path metadata reference to access the underlying metadata

        Calling this method returns a reference to an arbitrary piece of metadata associated
        with this path. Plugins can store any arbitrary data associated with a given path.
        Keep in mind that the metadata is stored in memory, so keep the size in moderation.

        To store / update the metadata, the path must be in read-write mode. However,
        already stored metadata remains accessible even if the path becomes read-only.

        Note this method is not applicable if the path is detached

        :param metadata_type: Type of the metadata being stored.
        :return: A reference to the metadata.
        """
        raise NotImplementedError


class FlushableSubstvars(Substvars):
    __slots__ = ()

    @contextlib.contextmanager
    def flush(self) -> Iterator[str]:
        """Temporarily write the substvars to a file and then re-read it again

        >>> s = FlushableSubstvars()
        >>> 'Test:Var' in s
        False
        >>> with s.flush() as name, open(name, 'wt', encoding='utf-8') as fobj:
        ...     _ = fobj.write('Test:Var=bar\\n')  # "_ = " is to ignore the return value of write
        >>> 'Test:Var' in s
        True

        Used as a context manager to define when the file is flushed and can be
        accessed via the file system. If the context terminates successfully, the
        file is read and its content replaces the current substvars.

        This is mostly useful if the plugin needs to interface with a third-party
        tool that requires a file as interprocess communication (IPC) for sharing
        the substvars.

        The file may be truncated or completed replaced (change inode) as long as
        the provided path points to a regular file when the context manager
        terminates successfully.

        Note that any manipulation of the substvars via the `Substvars` API while
        the file is flushed will silently be discarded if the context manager completes
        successfully.
        """
        with tempfile.NamedTemporaryFile(mode="w+t", encoding="utf-8") as tmp:
            self.write_substvars(tmp)
            tmp.flush()  # Temping to use close, but then we have to manually delete the file.
            yield tmp.name
            # Re-open; seek did not work when I last tried (if I did it work, feel free to
            # convert back to seek - as long as it works!)
            with open(tmp.name, "rt", encoding="utf-8") as fd:
                self.read_substvars(fd)

    def save(self) -> None:
        # Promote the debputy extension over `save()` for the plugins.
        if self._substvars_path is None:
            raise TypeError(
                "Please use `flush()` extension to temporarily write the substvars to the file system"
            )
        super().save()


class ServiceRegistry(Generic[DSD]):
    __slots__ = ()

    def register_service(
        self,
        path: VirtualPath,
        name: Union[str, List[str]],
        *,
        type_of_service: str = "service",  # "timer", etc.
        service_scope: str = "system",
        enable_by_default: bool = True,
        start_by_default: bool = True,
        default_upgrade_rule: ServiceUpgradeRule = "restart",
        service_context: Optional[DSD] = None,
    ) -> None:
        """Register a service detected in the package

        All the details will either be provided as-is or used as default when the plugin provided
        integration code is called.

        Two services from different service managers are considered related when:

         1) They are of the same type (`type_of_service`) and has the same scope (`service_scope`), AND
         2) Their plugin provided names has an overlap

        Related services can be covered by the same service definition in the manifest.

        :param path: The path defining this service.
        :param name: The name of the service. Multiple ones can be provided if the service has aliases.
          Note that when providing multiple names, `debputy` will use the first name in the list as the
          default name if it has to choose. Any alternative name provided can be used by the packager
          to identify this service.
        :param type_of_service: The type of service. By default, this is "service", but plugins can
          provide other types (such as "timer" for the systemd timer unit).
        :param service_scope: The scope for this service. By default, this is "system" meaning the
          service is a system-wide service. Service managers can define their own scopes such as
          "user" (which is used by systemd for "per-user" services).
        :param enable_by_default: Whether the service should be enabled by default, assuming the
          packager does not explicitly override this setting.
        :param start_by_default: Whether the service should be started by default on install, assuming
          the packager does not explicitly override this setting.
        :param default_upgrade_rule: The default value for how the service should be processed during
          upgrades. Options are:
              * `do-nothing`: The plugin should not interact with the running service (if any)
                (maintenance of the enabled start, start on install, etc. are still applicable)
              * `reload`: The plugin should attempt to reload the running service (if any).
                 Note: In combination with `auto_start_in_install == False`, be careful to not
                 start the service if not is not already running.
              * `restart`: The plugin should attempt to restart the running service (if any).
                 Note: In combination with `auto_start_in_install == False`, be careful to not
                 start the service if not is not already running.
              * `stop-then-start`: The plugin should stop the service during `prerm upgrade`
                 and start it against in the `postinst` script.

        :param service_context: Any custom data that the detector want to pass along to the
          integrator for this service.
        """
        raise NotImplementedError


@dataclasses.dataclass(slots=True, frozen=True)
class ParserAttributeDocumentation:
    attributes: FrozenSet[str]
    description: Optional[str]

    @property
    def is_hidden(self) -> bool:
        return False


@final
@dataclasses.dataclass(slots=True, frozen=True)
class StandardParserAttributeDocumentation(ParserAttributeDocumentation):
    sort_category: int = 0


def undocumented_attr(attr: str) -> ParserAttributeDocumentation:
    """Describe an attribute as undocumented

    If you for some reason do not want to document a particular attribute, you can mark it as
    undocumented. This is required if you are only documenting a subset of the attributes,
    because `debputy` assumes any omission to be a mistake.

    :param attr: Name of the attribute
    """
    return ParserAttributeDocumentation(
        frozenset({attr}),
        None,
    )


@dataclasses.dataclass(slots=True, frozen=True)
class ParserDocumentation:
    synopsis: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    attribute_doc: Optional[Sequence[ParserAttributeDocumentation]] = None
    alt_parser_description: Optional[str] = None
    documentation_reference_url: Optional[str] = None

    def replace(self, **changes: Any) -> "ParserDocumentation":
        return dataclasses.replace(self, **changes)

    @classmethod
    def from_ref_doc(cls, ref_doc: "ParserRefDocumentation") -> "ParserDocumentation":
        attr = [
            documented_attr(d["attr"], d["description"])
            for d in ref_doc.get("attributes", [])
        ]
        undoc_attr = ref_doc.get("undocumented_attributes")
        if undoc_attr:
            attr.extend(undocumented_attr(attr) for attr in undoc_attr)

        return reference_documentation(
            title=ref_doc["title"],
            description=ref_doc["description"],
            attributes=attr,
            non_mapping_description=ref_doc.get("non_mapping_description"),
            reference_documentation_url=ref_doc.get("ref_doc_url"),
            synopsis=ref_doc.get("synopsis"),
        )


@dataclasses.dataclass(slots=True, frozen=True)
class TypeMappingExample(Generic[S]):
    source_input: S


@dataclasses.dataclass(slots=True, frozen=True)
class TypeMappingDocumentation(Generic[S]):
    description: Optional[str] = None
    examples: Sequence[TypeMappingExample[S]] = tuple()


def type_mapping_example(source_input: S) -> TypeMappingExample[S]:
    return TypeMappingExample(source_input)


def type_mapping_reference_documentation(
    *,
    description: Optional[str] = None,
    examples: Union[TypeMappingExample[S], Iterable[TypeMappingExample[S]]] = tuple(),
) -> TypeMappingDocumentation[S]:
    e = (
        tuple([examples])
        if isinstance(examples, TypeMappingExample)
        else tuple(examples)
    )
    return TypeMappingDocumentation(
        description=description,
        examples=e,
    )


def documented_attr(
    attr: Union[str, Iterable[str]],
    description: str,
) -> ParserAttributeDocumentation:
    """Describe an attribute or a group of attributes

    :param attr: A single attribute or a sequence of attributes. The attribute must be the
      attribute name as used in the source format version of the TypedDict.

      If multiple attributes are provided, they will be documented together. This is often
      useful if these attributes are strongly related (such as different names for the same
      target attribute).
    :param description: The description the user should see for this attribute / these
       attributes. This parameter can be a Python format string with variables listed in
       the description of `reference_documentation`.
    :return: An opaque representation of the documentation,
    """
    attributes = [attr] if isinstance(attr, str) else attr
    return ParserAttributeDocumentation(
        frozenset(attributes),
        description,
    )


def reference_documentation(
    title: str = "Auto-generated reference documentation for ${RULE_NAME}",
    description: Optional[str] = textwrap.dedent(
        """\
            This is an automatically generated reference documentation for ${RULE_NAME}. It is generated
            from input provided by ${PLUGIN_NAME} via the debputy API.

            (If you are the provider of the ${PLUGIN_NAME} plugin, you can replace this text with
             your own documentation by providing the `inline_reference_documentation` when registering
             the manifest rule.)
            """
    ),
    attributes: Optional[Sequence[ParserAttributeDocumentation]] = None,
    non_mapping_description: Optional[str] = None,
    reference_documentation_url: Optional[str] = None,
    synopsis: Optional[str] = None,
) -> ParserDocumentation:
    """Provide inline reference documentation for the manifest snippet

    For parameters that mention that they are a Python format, the following template variables
    are available (`${FOO}`):

     * RULE_NAME: Name of the rule. If manifest snippet has aliases, this will be the name of
       the alias provided by the user.
     * MANIFEST_FORMAT_DOC: Path OR URL to the "MANIFEST-FORMAT" reference documentation from
       `debputy`. By using the MANIFEST_FORMAT_DOC variable, you ensure that you point to the
       file that matches the version of `debputy` itself.
     * PLUGIN_NAME: Name of the plugin providing this rule.

    :param title: The text you want the user to see as for your rule. A placeholder is provided by default.
      This parameter can be a Python format string with the above listed variables.
    :param description: The text you want the user to see as a description for the rule. An auto-generated
      placeholder is provided by default saying that no human written documentation was provided.
      This parameter can be a Python format string with the above listed variables.
    :param synopsis: One-line plain-text description used for describing the feature during completion.
    :param attributes: A sequence of attribute-related documentation. Each element of the sequence should
      be the result of `documented_attr` or `undocumented_attr`. The sequence must cover all source
      attributes exactly once.
    :param non_mapping_description: The text you want the user to see as the description for your rule when
      `debputy` describes its non-mapping format. Must not be provided for rules that do not have an
      (optional) non-mapping format as source format.  This parameter can be a Python format string with
      the above listed variables.
    :param reference_documentation_url: A URL to the reference documentation.
    :return: An opaque representation of the documentation,
    """
    return ParserDocumentation(
        synopsis,
        title,
        description,
        attributes,
        non_mapping_description,
        reference_documentation_url,
    )


class ServiceDefinition(Generic[DSD]):
    __slots__ = ()

    @property
    def name(self) -> str:
        """Name of the service registered by the plugin

        This is always a plugin provided name for this service (that is, `x.name in x.names`
        will always be `True`).  Where possible, this will be the same as the one that the
        packager provided when they provided any configuration related to this service.
        When not possible, this will be the first name provided by the plugin (`x.names[0]`).

        If all the aliases are equal, then using this attribute will provide traceability
        between the manifest and the generated maintscript snippets. When the exact name
        used is important, the plugin should ignore this attribute and pick the name that
        is needed.
        """
        raise NotImplementedError

    @property
    def names(self) -> Sequence[str]:
        """All *plugin provided* names and aliases of the service

        This is the name/sequence of names that the plugin provided when it registered
        the service earlier.
        """
        raise NotImplementedError

    @property
    def path(self) -> VirtualPath:
        """The registered path for this service

        :return: The path that was associated with this service when it was registered
          earlier.
        """
        raise NotImplementedError

    @property
    def type_of_service(self) -> str:
        """Type of the service such as "service" (daemon), "timer", etc.

        :return: The type of service scope. It is the same value as the one as the plugin provided
           when registering the service (if not explicitly provided, it defaults to "service").
        """
        raise NotImplementedError

    @property
    def service_scope(self) -> str:
        """Service scope such as "system" or "user"

        :return: The service scope. It is the same value as the one as the plugin provided
           when registering the service (if not explicitly provided, it defaults to "system")
        """
        raise NotImplementedError

    @property
    def auto_enable_on_install(self) -> bool:
        """Whether the service should be auto-enabled on install

        :return: True if the service should be enabled automatically, false if not.
        """
        raise NotImplementedError

    @property
    def auto_start_on_install(self) -> bool:
        """Whether the service should be auto-started on install

        :return: True if the service should be started automatically, false if not.
        """
        raise NotImplementedError

    @property
    def on_upgrade(self) -> ServiceUpgradeRule:
        """How to handle the service during an upgrade

        Options are:
          * `do-nothing`: The plugin should not interact with the running service (if any)
            (maintenance of the enabled start, start on install, etc. are still applicable)
          * `reload`: The plugin should attempt to reload the running service (if any).
             Note: In combination with `auto_start_in_install == False`, be careful to not
             start the service if not is not already running.
          * `restart`: The plugin should attempt to restart the running service (if any).
             Note: In combination with `auto_start_in_install == False`, be careful to not
             start the service if not is not already running.
          * `stop-then-start`: The plugin should stop the service during `prerm upgrade`
             and start it against in the `postinst` script.

        Note: In all cases, the plugin should still consider what to do in
        `prerm remove`, which is the last point in time where the plugin can rely on the
        service definitions in the file systems to stop the services when the package is
        being uninstalled.

        :return: The service restart rule
        """
        raise NotImplementedError

    @property
    def definition_source(self) -> str:
        """Describes where this definition came from

        If the definition is provided by the packager, then this will reference the part
        of the manifest that made this definition. Otherwise, this will be a reference
        to the plugin providing this definition.

        :return: The source of this definition
        """
        raise NotImplementedError

    @property
    def is_plugin_provided_definition(self) -> bool:
        """Whether the definition source points to the plugin or a package provided definition

        :return: True if definition is 100% from the plugin. False if the definition is partially
          or fully from another source (usually, the packager via the manifest).
        """
        raise NotImplementedError

    @property
    def service_context(self) -> Optional[DSD]:
        """Custom service context (if any) provided by the detector code of the plugin

        :return: If the detection code provided a custom data when registering the
          service, this attribute will reference that data.  If nothing was provided,
          then this attribute will be None.
        """
        raise NotImplementedError



import textwrap
from typing import Iterable, TYPE_CHECKING

from debputy.util import _info

if TYPE_CHECKING:
    from junit_xml import TestSuite, TestCase, to_xml_report_file
else:
    try:
        from junit_xml import TestSuite, TestCase, to_xml_report_file
    except ImportError:
        pass


from debputy.linting.lint_util import (
    LintReport,
    LintDiagnosticResult,
    LintDiagnosticResultState,
    debputy_severity,
)


class JunitLintReport(LintReport):

    def __init__(self, output_filename: str) -> None:
        super().__init__()
        self._output_filename = output_filename

    def finish_report(self) -> None:
        # Nothing to do for now
        all_test_cases = list(self._as_test_cases())
        test_suites = [
            TestSuite(
                "debputy lint",
                test_cases=all_test_cases,
                timestamp=str(self.start_timestamp),
            )
        ]
        with open(self._output_filename, "w", encoding="utf-8") as wfd:
            to_xml_report_file(wfd, test_suites, encoding="utf-8")
        _info(f"Wrote {self._output_filename}")

    def _as_test_cases(self) -> Iterable["TestCase"]:
        for filename, duration in self.durations.items():
            results = self.diagnostics_by_file.get(filename, [])
            yield self._as_test_case(filename, results, duration)

    def _as_test_case(
        self,
        filename: str,
        diagnostic_results: Iterable[LintDiagnosticResult],
        duration: float,
    ) -> "TestCase":
        if not duration:
            duration = 0.000001
        case = TestCase(
            filename,
            # The JUnit schema has `classname` as mandatory
            classname=filename,
            allow_multiple_subelements=True,
            elapsed_sec=duration,
        )
        for diagnostic_result in diagnostic_results:
            if diagnostic_result.result_state == LintDiagnosticResultState.FIXED:
                continue
            diagnostic = diagnostic_result.diagnostic
            severity = debputy_severity(diagnostic)
            if diagnostic_result.is_file_level_diagnostic:
                range_desc = "entire file"
            else:
                range_desc = str(diagnostic.range)
            code = f" [{diagnostic.code}]" if diagnostic.code else ""
            authority = diagnostic.source
            assert authority is not None
            output = textwrap.dedent(
                f"""\
            {filename}{code} ({severity}) {range_desc} [{authority}]: {diagnostic.message}
            """
            )
            case.add_failure_info(
                message=f"[{severity}]: {diagnostic.message}",
                output=output,
            )
        return case



from typing import List, Optional, Sequence, Union, Iterable, TYPE_CHECKING

from debputy.commands.debputy_cmd.output import MAN_URL_REWRITE
from debputy.lsprotocol.types import (
    TextEdit,
    Position,
    Range,
    WillSaveTextDocumentParams,
    DocumentFormattingParams,
)

from debputy.linting.lint_util import LinterPositionCodec

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass

try:
    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument, PositionCodec
except ImportError:
    pass

if TYPE_CHECKING:
    LintCapablePositionCodec = Union[LinterPositionCodec, PositionCodec]
else:
    LintCapablePositionCodec = LinterPositionCodec


def markdown_urlify(uri: str) -> str:
    if uri.startswith("man:"):
        m = MAN_URL_REWRITE.match(uri)
        if m:
            page, section = m.groups()
            link_url = f"https://manpages.debian.org/{page}.{section}"
            return f"[{uri}]({link_url})"
        return uri

    return f"<{uri}>"


def normalize_dctrl_field_name(f: str) -> str:
    if not f or not f.startswith(("x", "X")):
        return f
    i = 0
    for i in range(1, len(f)):
        if f[i] == "-":
            i += 1
            break
        if f[i] not in ("b", "B", "s", "S", "c", "C"):
            return f
    assert i > 0
    return f[i:]


def on_save_trim_end_of_line_whitespace(
    ls: "LanguageServer",
    params: Union[WillSaveTextDocumentParams, DocumentFormattingParams],
) -> Optional[Sequence[TextEdit]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    return trim_end_of_line_whitespace(doc.position_codec, doc.lines)


def trim_end_of_line_whitespace(
    position_codec: "LintCapablePositionCodec",
    lines: List[str],
    *,
    line_range: Optional[Iterable[int]] = None,
    line_relative_line_no: int = 0,
) -> Optional[Sequence[TextEdit]]:
    edits = []
    if line_range is None:
        line_range = range(0, len(lines))
    for line_no in line_range:
        orig_line = lines[line_no]
        orig_len = len(orig_line)
        if orig_line.endswith("\n"):
            orig_len -= 1
        stripped_len = len(orig_line.rstrip())
        if stripped_len == orig_len:
            continue

        stripped_len_client_off = position_codec.client_num_units(
            orig_line[:stripped_len]
        )
        orig_len_client_off = position_codec.client_num_units(orig_line[:orig_len])
        edit_range = position_codec.range_to_client_units(
            lines,
            Range(
                Position(
                    line_no + line_relative_line_no,
                    stripped_len_client_off,
                ),
                Position(
                    line_no + line_relative_line_no,
                    orig_len_client_off,
                ),
            ),
        )
        edits.append(
            TextEdit(
                edit_range,
                "",
            )
        )

    return edits


class SemanticTokensState:
    __slots__ = ("ls", "doc", "lines", "tokens", "_previous_line", "_previous_col")

    def __init__(
        self,
        ls: "DebputyLanguageServer",
        doc: "TextDocument",
        lines: List[str],
        tokens: List[int],
    ) -> None:
        self.ls = ls
        self.doc = doc
        self.lines = lines
        self.tokens = tokens
        self._previous_line = 0
        self._previous_col = 0

    def emit_token(
        self,
        start_pos: Position,
        len_client_units: int,
        token_code: int,
        *,
        token_modifiers: int = 0,
    ) -> None:
        line_delta = start_pos.line - self._previous_line
        self._previous_line = start_pos.line
        previous_col = self._previous_col

        if line_delta:
            previous_col = 0

        column_delta = start_pos.character - previous_col
        self._previous_col = start_pos.character

        tokens = self.tokens
        tokens.append(line_delta)  # Line delta
        tokens.append(column_delta)  # Token column delta
        tokens.append(len_client_units)  # Token length
        tokens.append(token_code)
        tokens.append(token_modifiers)



import asyncio
import collections
from typing import (
    List,
    Optional,
    Mapping,
    Any,
    Callable,
    Sequence,
    TYPE_CHECKING,
    Awaitable,
)

import pytest

from debputy.filesystem_scan import VirtualPathBase
from debputy.linting.lint_util import (
    LinterPositionCodec,
    LintStateImpl,
    LintState,
)
from debputy.lsp.maint_prefs import (
    MaintainerPreferenceTable,
    EffectiveFormattingPreference,
)
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet

if TYPE_CHECKING:
    from lsprotocol import types
else:
    from debputy.lsprotocol import types


try:
    from Levenshtein import distance

    HAS_LEVENSHTEIN = True
except ImportError:
    HAS_LEVENSHTEIN = False

LINTER_POSITION_CODEC = LinterPositionCodec()


class LintWrapper:

    def __init__(
        self,
        path: str,
        handler: Callable[[LintState], Awaitable[None]],
        debputy_plugin_feature_set: PluginProvidedFeatureSet,
        dctrl_parser: DctrlParser,
    ) -> None:
        self._debputy_plugin_feature_set = debputy_plugin_feature_set
        self._handler = handler
        self.dctrl_lines: Optional[List[str]] = None
        self.path = path
        self._dctrl_parser = dctrl_parser
        self.source_root: Optional[VirtualPathBase] = None
        self.lint_maint_preference_table = MaintainerPreferenceTable({}, {})
        self.effective_preference: Optional[EffectiveFormattingPreference] = None

    def __call__(self, lines: List[str]) -> List["types.Diagnostic"]:
        source_package = None
        binary_packages = None
        dctrl_lines = self.dctrl_lines
        if dctrl_lines is not None:
            _, source_package, binary_packages = (
                self._dctrl_parser.parse_source_debian_control(
                    dctrl_lines, ignore_errors=True
                )
            )
        source_root = self.source_root
        debian_dir = source_root.get("debian") if source_root is not None else None
        state = LintStateImpl(
            self._debputy_plugin_feature_set,
            self.lint_maint_preference_table,
            source_root,
            debian_dir,
            self.path,
            "".join(dctrl_lines) if dctrl_lines is not None else "",
            lines,
            source_package,
            binary_packages,
            self.effective_preference,
            self._handler,
        )
        return check_diagnostics(asyncio.run(state.gather_diagnostics()))


def requires_levenshtein(func: Any) -> Any:
    return pytest.mark.skipif(
        not HAS_LEVENSHTEIN, reason="Missing python3-levenshtein"
    )(func)


def check_diagnostics(
    diagnostics: Optional[List["types.Diagnostic"]],
) -> List["types.Diagnostic"]:
    if diagnostics:
        for diagnostic in diagnostics:
            assert diagnostic.severity is not None
            assert diagnostic.source is not None
    elif diagnostics is None:
        diagnostics = []
    return diagnostics


def by_range_sort_key(diagnostic: types.Diagnostic) -> Any:
    start_pos = diagnostic.range.start
    end_pos = diagnostic.range.end
    return start_pos.line, start_pos.character, end_pos.line, end_pos.character


def group_diagnostics_by_severity(
    diagnostics: Optional[List["types.Diagnostic"]],
) -> Mapping["types.DiagnosticSeverity", List["types.Diagnostic"]]:
    if not diagnostics:
        return {}

    by_severity = collections.defaultdict(list)

    for diagnostic in sorted(diagnostics, key=by_range_sort_key):
        severity = diagnostic.severity
        assert severity is not None
        by_severity[severity].append(diagnostic)

    return by_severity


def diag_range_to_text(lines: Sequence[str], range_: "types.Range") -> str:
    parts = []
    for line_no in range(range_.start.line, range_.end.line + 1):
        line = lines[line_no]
        chunk = line
        if line_no == range_.start.line and line_no == range_.end.line:
            chunk = line[range_.start.character : range_.end.character]
        elif line_no == range_.start.line:
            chunk = line[range_.start.character :]
        elif line_no == range_.end.line:
            chunk = line[: range_.end.character]
        parts.append(chunk)
    return "".join(parts)
